# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2013 OpenStack Foundation
# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import babel.dates
import datetime
import dateutil.parser
import dateutil.tz
import hashlib
import json
import jwt
import logging
import prettytable
import os
import re
import sys
import time
import textwrap
import requests
import urllib.parse

import zuul.cmd
from zuul.lib.config import get_default
from zuul.model import (
    Abide,
    ImageBuildArtifact,
    ImageUpload,
    Pipeline,
    PipelineChangeList,
    PipelineState,
    SystemAttributes,
)
from zuul.zk import ZooKeeperClient
from zuul.lib.keystorage import KeyStorage
from zuul.manager.independent import IndependentPipelineManager
from zuul.zk.locks import tenant_read_lock, pipeline_lock
from zuul.zk.zkobject import ZKContext
from zuul.zk.components import COMPONENT_REGISTRY
from zuul.zk.image_registry import (
    ImageBuildRegistry,
    ImageUploadRegistry,
)

from kazoo.exceptions import NoNodeError, NodeExistsError


def parse_cutoff(now, before, older_than):
    if before and not older_than:
        cutoff = dateutil.parser.parse(before)
        if cutoff.tzinfo and cutoff.tzinfo != dateutil.tz.tzutc():
            raise RuntimeError("Timestamp must be specified as UTC")
        cutoff = cutoff.replace(tzinfo=dateutil.tz.tzutc())
        return cutoff
    elif older_than and not before:
        value = older_than[:-1]
        suffix = older_than[-1]
        if suffix == 'd':
            delta = datetime.timedelta(days=int(value))
        elif suffix == 'h':
            delta = datetime.timedelta(hours=int(value))
        else:
            raise RuntimeError("Unsupported relative time")
        return now - delta
    else:
        raise RuntimeError(
            "Either --before or --older-than must be supplied")


# todo This should probably live somewhere else
class ZuulRESTClient(object):
    """Basic client for Zuul's REST API"""
    def __init__(self, url, verify=False, auth_token=None):
        self.url = url
        if not self.url.endswith('/'):
            self.url += '/'
        self.auth_token = auth_token
        self.verify = verify
        self.base_url = urllib.parse.urljoin(self.url, 'api/')
        self.session = requests.Session()
        self.session.verify = self.verify
        self.session.headers = dict(
            Authorization='Bearer %s' % self.auth_token)

    def _check_status(self, req):
        try:
            req.raise_for_status()
        except Exception as e:
            if req.status_code == 401:
                print('Unauthorized - your token might be invalid or expired.')
            elif req.status_code == 403:
                print('Insufficient privileges to perform the action.')
            else:
                print('Unknown error: "%e"' % e)

    def autohold(self, tenant, project, job, change, ref,
                 reason, count, node_hold_expiration):
        if not self.auth_token:
            raise Exception('Auth Token required')
        args = {"reason": reason,
                "count": count,
                "job": job,
                "change": change,
                "ref": ref,
                "node_hold_expiration": node_hold_expiration}
        url = urllib.parse.urljoin(
            self.base_url,
            'tenant/%s/project/%s/autohold' % (tenant, project))
        req = self.session.post(url, json=args)
        self._check_status(req)
        return req.json()

    def autohold_list(self, tenant):
        if not tenant:
            raise Exception('"--tenant" argument required')
        url = urllib.parse.urljoin(
            self.base_url,
            'tenant/%s/autohold' % tenant)
        req = requests.get(url, verify=self.verify)
        self._check_status(req)
        return req.json()

    def enqueue(self, tenant, pipeline, project, trigger, change):
        if not self.auth_token:
            raise Exception('Auth Token required')
        args = {"trigger": trigger,
                "change": change,
                "pipeline": pipeline}
        url = urllib.parse.urljoin(
            self.base_url,
            'tenant/%s/project/%s/enqueue' % (tenant, project))
        req = self.session.post(url, json=args)
        self._check_status(req)
        return req.json()

    def enqueue_ref(self, tenant, pipeline, project,
                    trigger, ref, oldrev, newrev):
        if not self.auth_token:
            raise Exception('Auth Token required')
        args = {"trigger": trigger,
                "ref": ref,
                "oldrev": oldrev,
                "newrev": newrev,
                "pipeline": pipeline}
        url = urllib.parse.urljoin(
            self.base_url,
            'tenant/%s/project/%s/enqueue' % (tenant, project))
        req = self.session.post(url, json=args)
        self._check_status(req)
        return req.json()

    def dequeue(self, tenant, pipeline, project, change=None, ref=None):
        if not self.auth_token:
            raise Exception('Auth Token required')
        args = {"pipeline": pipeline}
        if change and not ref:
            args['change'] = change
        elif ref and not change:
            args['ref'] = ref
        else:
            raise Exception('need change OR ref')
        url = urllib.parse.urljoin(
            self.base_url,
            'tenant/%s/project/%s/dequeue' % (tenant, project))
        req = self.session.post(url, json=args)
        self._check_status(req)
        return req.json()

    def promote(self, tenant, pipeline, change_ids):
        if not self.auth_token:
            raise Exception('Auth Token required')
        args = {
            "pipeline": pipeline,
            "changes": change_ids,
        }
        url = urllib.parse.urljoin(
            self.base_url,
            'tenant/%s/promote' % tenant)
        req = self.session.post(url, json=args)
        self._check_status(req)
        return req.json()

    def get_running_jobs(self, *args, **kwargs):
        raise NotImplementedError(
            'This action is unsupported by the REST API')


class Client(zuul.cmd.ZuulApp):
    app_name = 'zuul'
    app_description = 'Zuul CLI client.'
    log = logging.getLogger("zuul.Client")

    def createParser(self):
        parser = super(Client, self).createParser()
        parser.add_argument('-v', dest='verbose', action='store_true',
                            help='verbose output')
        parser.add_argument('--auth-token', dest='auth_token',
                            required=False,
                            default=None,
                            help='Authentication Token, needed if using the'
                                 'REST API')
        parser.add_argument('--zuul-url', dest='zuul_url',
                            required=False,
                            default=None,
                            help='Zuul API URL, needed if using the '
                                 'REST API without a configuration file')
        parser.add_argument('--insecure', dest='insecure_ssl',
                            required=False,
                            action='store_false',
                            help='Do not verify SSL connection to Zuul, '
                                 'when using the REST API (Defaults to False)')

        subparsers = parser.add_subparsers(title='commands',
                                           description='valid commands',
                                           help='additional help')

        # Autohold
        cmd_autohold = subparsers.add_parser(
            'autohold', help='[DEPRECATED - use zuul-client] '
                             'hold nodes for failed job')
        cmd_autohold.add_argument('--tenant', help='tenant name',
                                  required=True)
        cmd_autohold.add_argument('--project', help='project name',
                                  required=True)
        cmd_autohold.add_argument('--job', help='job name',
                                  required=True)
        cmd_autohold.add_argument('--change',
                                  help='specific change to hold nodes for',
                                  required=False, default='')
        cmd_autohold.add_argument('--ref', help='git ref to hold nodes for',
                                  required=False, default='')
        cmd_autohold.add_argument('--reason', help='reason for the hold',
                                  required=True)
        cmd_autohold.add_argument('--count',
                                  help='number of job runs (default: 1)',
                                  required=False, type=int, default=1)
        cmd_autohold.add_argument(
            '--node-hold-expiration',
            help=('how long in seconds should the node set be in HOLD status '
                  '(default: scheduler\'s default_hold_expiration value)'),
            required=False, type=int)
        cmd_autohold.set_defaults(func=self.autohold)

        cmd_autohold_delete = subparsers.add_parser(
            'autohold-delete', help='[DEPRECATED - use zuul-client] '
                                    'delete autohold request')
        cmd_autohold_delete.set_defaults(func=self.autohold_delete)
        cmd_autohold_delete.add_argument('id', metavar='REQUEST_ID',
                                         help='the hold request ID')

        cmd_autohold_info = subparsers.add_parser(
            'autohold-info', help='[DEPRECATED - use zuul-client] '
                                  'retrieve autohold request detailed info')
        cmd_autohold_info.set_defaults(func=self.autohold_info)
        cmd_autohold_info.add_argument('id', metavar='REQUEST_ID',
                                       help='the hold request ID')

        cmd_autohold_list = subparsers.add_parser(
            'autohold-list', help='[DEPRECATED - use zuul-client] '
                                  'list autohold requests')
        cmd_autohold_list.add_argument('--tenant', help='tenant name',
                                       required=True)
        cmd_autohold_list.set_defaults(func=self.autohold_list)

        # Enqueue/Dequeue
        cmd_enqueue = subparsers.add_parser(
            'enqueue',
            help='[DEPRECATED - use zuul-client] enqueue a change')
        cmd_enqueue.add_argument('--tenant', help='tenant name',
                                 required=True)
        # TODO(mhu) remove in a few releases
        cmd_enqueue.add_argument('--trigger',
                                 help='trigger name (deprecated and ignored. '
                                      'Kept only for backward compatibility)',
                                 required=False, default=None)
        cmd_enqueue.add_argument('--pipeline', help='pipeline name',
                                 required=True)
        cmd_enqueue.add_argument('--project', help='project name',
                                 required=True)
        cmd_enqueue.add_argument('--change', help='change id',
                                 required=True)
        cmd_enqueue.set_defaults(func=self.enqueue)

        cmd_enqueue = subparsers.add_parser(
            'enqueue-ref',
            help='[DEPRECATED - use zuul-client] enqueue a ref',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Submit a trigger event

            Directly enqueue a trigger event.  This is usually used
            to manually "replay" a trigger received from an external
            source such as gerrit.'''))
        cmd_enqueue.add_argument('--tenant', help='tenant name',
                                 required=True)
        cmd_enqueue.add_argument('--trigger', help='trigger name',
                                 required=False, default=None)
        cmd_enqueue.add_argument('--pipeline', help='pipeline name',
                                 required=True)
        cmd_enqueue.add_argument('--project', help='project name',
                                 required=True)
        cmd_enqueue.add_argument('--ref', help='ref name',
                                 required=True)
        cmd_enqueue.add_argument(
            '--oldrev', help='old revision', default=None)
        cmd_enqueue.add_argument(
            '--newrev', help='new revision', default=None)
        cmd_enqueue.set_defaults(func=self.enqueue_ref)

        cmd_dequeue = subparsers.add_parser(
            'dequeue',
            help='[DEPRECATED - use zuul-client] '
                 'dequeue a buildset by its '
                 'change or ref')
        cmd_dequeue.add_argument('--tenant', help='tenant name',
                                 required=True)
        cmd_dequeue.add_argument('--pipeline', help='pipeline name',
                                 required=True)
        cmd_dequeue.add_argument('--project', help='project name',
                                 required=True)
        cmd_dequeue.add_argument('--change', help='change id',
                                 default=None)
        cmd_dequeue.add_argument('--ref', help='ref name',
                                 default=None)
        cmd_dequeue.set_defaults(func=self.dequeue)

        # Promote
        cmd_promote = subparsers.add_parser(
            'promote',
            help='[DEPRECATED - use zuul-client] '
                 'promote one or more changes')
        cmd_promote.add_argument('--tenant', help='tenant name',
                                 required=True)
        cmd_promote.add_argument('--pipeline', help='pipeline name',
                                 required=True)
        cmd_promote.add_argument('--changes', help='change ids',
                                 required=True, nargs='+')
        cmd_promote.set_defaults(func=self.promote)

        # Show
        cmd_show = subparsers.add_parser(
            'show',
            help='[DEPRECATED - use zuul-client] '
                 'show current statuses')
        cmd_show.set_defaults(func=self.show_running_jobs)
        show_subparsers = cmd_show.add_subparsers(title='show')
        show_running_jobs = show_subparsers.add_parser(
            'running-jobs',
            help='show the running jobs'
        )
        running_jobs_columns = list(self._show_running_jobs_columns().keys())
        show_running_jobs.add_argument(
            '--columns',
            help="comma separated list of columns to display (or 'ALL')",
            choices=running_jobs_columns.append('ALL'),
            default='name, worker.name, start_time, result'
        )

        # TODO: add filters such as queue, project, changeid etc
        show_running_jobs.set_defaults(func=self.show_running_jobs)

        # Conf check
        cmd_conf_check = subparsers.add_parser(
            'tenant-conf-check',
            help='validate the tenant configuration')
        cmd_conf_check.set_defaults(func=self.validate)

        # Auth token
        cmd_create_auth_token = subparsers.add_parser(
            'create-auth-token',
            help='create an Authentication Token for the web API',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Create an Authentication Token for the administration web API

            Create a bearer token that can be used to access Zuul's
            administration web API. This is typically used to delegate
            privileged actions such as enqueueing and autoholding to
            third parties, scoped to a single tenant.
            At least one authenticator must be configured with a secret
            that can be used to sign the token.'''))
        cmd_create_auth_token.add_argument(
            '--auth-config',
            help=('The authenticator to use. '
                  'Must match an authenticator defined in zuul\'s '
                  'configuration file.'),
            default='zuul_operator',
            required=True)
        cmd_create_auth_token.add_argument(
            '--tenant',
            help=('When specified, zuul.admin claim with the value '
                  'of the tenant will be added to the token.'),
            required=False)
        cmd_create_auth_token.add_argument(
            '--user',
            help=("The user's name. Used for traceability in logs."),
            default=None,
            required=True)
        cmd_create_auth_token.add_argument(
            '--expires-in',
            help=('Token validity duration in seconds '
                  '(default: %i)' % 600),
            type=int,
            default=600,
            required=False)
        cmd_create_auth_token.add_argument(
            '--claim',
            help=('Custom claim in format claim:value. It can '
                  'be used multiple times to add multiple claims.'),
            action="extend",
            type=lambda x: tuple(x.split(':')),
            nargs='*',
            default=[])
        cmd_create_auth_token.add_argument(
            '--print-meta-info',
            action='store_true',
            help="When specified, the meta info of the token will be printed")
        cmd_create_auth_token.set_defaults(func=self.create_auth_token)

        # Key storage
        cmd_import_keys = subparsers.add_parser(
            'import-keys',
            help='import project keys to ZooKeeper',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Import previously exported project secret keys to ZooKeeper

            Given a file with previously exported project keys, this
            command will import them into ZooKeeper.  Existing keys
            will not be overwritten; to overwrite keys, add the
            --force flag.'''))
        cmd_import_keys.set_defaults(command='import-keys')
        cmd_import_keys.add_argument('path', type=str,
                                     help='key export file path')
        cmd_import_keys.add_argument('--force', action='store_true',
                                     help='overwrite existing keys')
        cmd_import_keys.set_defaults(func=self.import_keys)

        cmd_export_keys = subparsers.add_parser(
            'export-keys',
            help='export project keys from ZooKeeper',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Export project secret keys from ZooKeeper

            This command exports project secret keys from ZooKeeper
            and writes them to a file which is suitable for backing
            up and later use with the import-keys command.

            The key contents are still protected by the keystore
            password and can not be used or decrypted without it.'''))
        cmd_export_keys.set_defaults(command='export-keys')
        cmd_export_keys.add_argument('path', type=str,
                                     help='key export file path')
        cmd_export_keys.set_defaults(func=self.export_keys)

        cmd_copy_keys = subparsers.add_parser(
            'copy-keys',
            help='copy keys from one project to another',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Copy secret keys from one project to another

            When projects are renamed, this command may be used to
            copy the secret keys from the current name to the new name
            in order to avoid service interruption.'''))
        cmd_copy_keys.set_defaults(command='copy-keys')
        cmd_copy_keys.add_argument('src_connection', type=str,
                                   help='original connection name')
        cmd_copy_keys.add_argument('src_project', type=str,
                                   help='original project name')
        cmd_copy_keys.add_argument('dest_connection', type=str,
                                   help='new connection name')
        cmd_copy_keys.add_argument('dest_project', type=str,
                                   help='new project name')
        cmd_copy_keys.set_defaults(func=self.copy_keys)

        cmd_delete_keys = subparsers.add_parser(
            'delete-keys',
            help='delete project keys',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Delete the ssh and secrets keys for a project
            '''))
        cmd_delete_keys.set_defaults(command='delete-keys')
        cmd_delete_keys.add_argument('connection', type=str,
                                     help='connection name')
        cmd_delete_keys.add_argument('project', type=str,
                                     help='project name')
        cmd_delete_keys.set_defaults(func=self.delete_keys)

        cmd_delete_oidc_signing_keys = subparsers.add_parser(
            'delete-oidc-signing-keys',
            help='delete OIDC signing keys',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Delete the OIDC signing keys of an algorithm
            '''))
        cmd_delete_oidc_signing_keys.set_defaults(
            command='delete-oidc-signing-keys')
        cmd_delete_oidc_signing_keys.add_argument(
            'algorithm', type=str, help='algorithm name')
        cmd_delete_oidc_signing_keys.set_defaults(
            func=self.delete_oidc_signing_keys)

        # image storage
        cmd_import_images = subparsers.add_parser(
            'import-images',
            help='import images to ZooKeeper',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Import previously exported images to ZooKeeper

            Given a file with previously exported images, this
            command will import them into ZooKeeper.  Existing images
            will not be overwritten.'''))
        cmd_import_images.set_defaults(command='import-images')
        cmd_import_images.add_argument('path', type=str,
                                       help='image export file path')
        cmd_import_images.set_defaults(func=self.import_images)

        cmd_export_images = subparsers.add_parser(
            'export-images',
            help='export images from ZooKeeper',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Export images from ZooKeeper

            This command exports images from ZooKeeper
            and writes them to a file which is suitable for backing
            up and later use with the import-images command.'''))
        cmd_export_images.set_defaults(command='export-images')
        cmd_export_images.add_argument('path', type=str,
                                       help='image export file path')
        cmd_export_images.set_defaults(func=self.export_images)

        # ZK Maintenance
        cmd_delete_state = subparsers.add_parser(
            'delete-state',
            help='delete ephemeral ZooKeeper state',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Delete all ephemeral state stored in ZooKeeper

            Zuul stores a considerable amount of ephemeral state
            information in ZooKeeper.  Generally it should be able to
            detect and correct any errors, but if the state becomes
            corrupted and it is unable to recover, this command may be
            used to delete all ephemeral data from ZooKeeper and start
            anew.

            Do not run this command while any Zuul component is
            running (perform a complete shutdown first).

            This command will only remove ephemeral Zuul data from
            ZooKeeper; it will not remove private keys or Nodepool
            data.'''))
        cmd_delete_state.set_defaults(command='delete-state')
        cmd_delete_state.add_argument(
            '--keep-config-cache', action='store_true',
            help='keep config cache')
        cmd_delete_state.set_defaults(func=self.delete_state)

        cmd_delete_pipeline_state = subparsers.add_parser(
            'delete-pipeline-state',
            help='delete single pipeline ZooKeeper state',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Delete a single pipeline state stored in ZooKeeper

            In the unlikely event that a bug in Zuul or ZooKeeper data
            corruption occurs in such a way that it only affects a
            single pipeline, this command might be useful in
            attempting to recover.

            The circumstances under which this command will be able to
            effect a recovery are very rare and even so it may not be
            sufficient.  In general, if an error occurs it is better
            to shut Zuul down and run "zuul delete-state".

            This command will lock the specified tenant and
            then completely delete the pipeline state.'''))
        cmd_delete_pipeline_state.set_defaults(command='delete-pipeline-state')
        cmd_delete_pipeline_state.set_defaults(func=self.delete_pipeline_state)
        cmd_delete_pipeline_state.add_argument('tenant', type=str,
                                               help='tenant name')
        cmd_delete_pipeline_state.add_argument('pipeline', type=str,
                                               help='pipeline name')

        # DB Maintenance
        cmd_prune_database = subparsers.add_parser(
            'prune-database',
            help='prune old database entries',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent('''\
            Prune old database entries

            This command will delete database entries older than the
            specified cutoff (which can be specified as either an
            absolute or relative time).'''))
        cmd_prune_database.set_defaults(command='prune-database')
        cmd_prune_database.add_argument(
            '--before',
            help='absolute timestamp (e.g., "2022-01-31 12:00:00")')
        cmd_prune_database.add_argument(
            '--older-than',
            help='relative time (e.g., "24h" or "180d")')
        cmd_prune_database.add_argument(
            '--batch-size',
            default=10000,
            help='transaction batch size')
        cmd_prune_database.set_defaults(func=self.prune_database)

        return parser

    def parseArguments(self, args=None):
        parser = super(Client, self).parseArguments()
        if not getattr(self.args, 'func', None):
            parser.print_help()
            sys.exit(1)
        if self.args.func == self.enqueue_ref:
            # if oldrev or newrev is set, ensure they're not the same
            if (self.args.oldrev is not None) or \
               (self.args.newrev is not None):
                if self.args.oldrev == self.args.newrev:
                    parser.error(
                        "The old and new revisions must not be the same.")
        if self.args.func == self.dequeue:
            if self.args.change is None and self.args.ref is None:
                parser.error("Change or ref needed.")
            if self.args.change is not None and self.args.ref is not None:
                parser.error(
                    "The 'change' and 'ref' arguments are mutually exclusive.")

    def setup_logging(self):
        """Client logging does not rely on conf file"""
        if self.args.verbose:
            logging.basicConfig(level=logging.DEBUG)

    def main(self):
        self.parseArguments()
        if not self.args.zuul_url:
            self.readConfig()
        self.setup_logging()
        if self.args.func in [self.autohold, self.autohold_delete,
                              self.enqueue, self.enqueue_ref,
                              self.dequeue, self.promote]:
            print(
                "Warning: this command is deprecated with zuul-admin, "
                "please use `zuul-client` instead",
                file=sys.stderr)
        if self.args.func():
            sys.exit(0)
        else:
            sys.exit(1)

    def get_client(self):
        if self.args.zuul_url:
            self.log.debug('Zuul URL provided as argument, using REST client')
            client = ZuulRESTClient(self.args.zuul_url,
                                    self.args.insecure_ssl,
                                    self.args.auth_token)
            return client
        conf_sections = self.config.sections()
        if 'webclient' in conf_sections:
            self.log.debug('web section found in config, using REST client')
            server = get_default(self.config, 'webclient', 'url', None)
            verify = get_default(self.config, 'webclient', 'verify_ssl',
                                 self.args.insecure_ssl)
            client = ZuulRESTClient(server, verify,
                                    self.args.auth_token)
        else:
            print('Unable to find a way to connect to Zuul, add a '
                  '"webclient" section to your configuration file')
            sys.exit(1)
        if server is None:
            print('Missing "server" configuration value')
            sys.exit(1)
        return client

    def autohold(self):
        if self.args.change and self.args.ref:
            print("Change and ref can't be both used for the same request")
            return False
        if "," in self.args.change:
            print("Error: change argument can not contain any ','")
            return False

        node_hold_expiration = self.args.node_hold_expiration
        client = self.get_client()
        r = client.autohold(
            tenant=self.args.tenant,
            project=self.args.project,
            job=self.args.job,
            change=self.args.change,
            ref=self.args.ref,
            reason=self.args.reason,
            count=self.args.count,
            node_hold_expiration=node_hold_expiration)
        return r

    def autohold_delete(self):
        client = self.get_client()
        return client.autohold_delete(self.args.id)

    def autohold_info(self):
        client = self.get_client()
        request = client.autohold_info(self.args.id)

        if not request:
            print("Autohold request not found")
            return True

        print("ID: %s" % request['id'])
        print("Tenant: %s" % request['tenant'])
        print("Project: %s" % request['project'])
        print("Job: %s" % request['job'])
        print("Ref Filter: %s" % request['ref_filter'])
        print("Max Count: %s" % request['max_count'])
        print("Current Count: %s" % request['current_count'])
        print("Node Expiration: %s" % request['node_expiration'])
        print("Request Expiration: %s" % time.ctime(request['expired']))
        print("Reason: %s" % request['reason'])
        print("Held Nodes: %s" % request['nodes'])

        return True

    def autohold_list(self):
        client = self.get_client()
        autohold_requests = client.autohold_list(tenant=self.args.tenant)

        if not autohold_requests:
            print("No autohold requests found")
            return True

        table = prettytable.PrettyTable(
            field_names=[
                'ID', 'Tenant', 'Project', 'Job', 'Ref Filter',
                'Current Count', 'Max Count', 'Reason'
            ])

        for request in autohold_requests:
            table.add_row([
                request['id'],
                request['tenant'],
                request['project'],
                request['job'],
                request['ref_filter'],
                request['current_count'],
                request['max_count'],
                request['reason'],
            ])

        print(table)
        return True

    def enqueue(self):
        client = self.get_client()
        r = client.enqueue(
            tenant=self.args.tenant,
            pipeline=self.args.pipeline,
            project=self.args.project,
            trigger=self.args.trigger,
            change=self.args.change)
        return r

    def enqueue_ref(self):
        client = self.get_client()
        r = client.enqueue_ref(
            tenant=self.args.tenant,
            pipeline=self.args.pipeline,
            project=self.args.project,
            trigger=self.args.trigger,
            ref=self.args.ref,
            oldrev=self.args.oldrev,
            newrev=self.args.newrev)
        return r

    def dequeue(self):
        client = self.get_client()
        r = client.dequeue(
            tenant=self.args.tenant,
            pipeline=self.args.pipeline,
            project=self.args.project,
            change=self.args.change,
            ref=self.args.ref)
        return r

    def create_auth_token(self):
        auth_section = ''
        for section_name in self.config.sections():
            if re.match(r'^auth ([\'\"]?)%s(\1)$' % self.args.auth_config,
                        section_name, re.I):
                auth_section = section_name
                break
        if auth_section == '':
            print('"%s" authenticator configuration not found.'
                  % self.args.auth_config)
            sys.exit(1)
        now = int(time.time())
        exp = now + self.args.expires_in
        token = {'iat': now,
                 'exp': exp,
                 'iss': get_default(self.config, auth_section, 'issuer_id'),
                 'aud': get_default(self.config, auth_section, 'client_id'),
                 'sub': self.args.user,
                }
        # Add zuul.admin claim only when tenant is specified
        if self.args.tenant:
            token['zuul'] = {'admin': [self.args.tenant, ]}

        # Add custom claims, allow to overwrite default claims
        # when it is required.
        for k, v in self.args.claim:
            token[k] = v

        driver = get_default(
            self.config, auth_section, 'driver')
        if driver == 'HS256':
            key = get_default(self.config, auth_section, 'secret')
        elif driver == 'RS256':
            private_key = get_default(self.config, auth_section, 'private_key')
            try:
                with open(private_key, 'r') as pk:
                    key = pk.read()
            except Exception as e:
                print('Could not read private key at "%s": %s' % (private_key,
                                                                  e))
                sys.exit(1)
        else:
            print('Unknown or unsupported authenticator driver "%s"' % driver)
            sys.exit(1)
        try:
            auth_token = jwt.encode(token,
                                    key=key,
                                    algorithm=driver)
            print("Bearer %s" % auth_token)
            if self.args.print_meta_info:
                print("---------------- Meta Info ----------------")
                if self.args.tenant:
                    print("Tenant:\t\t%s" % self.args.tenant)
                print("User:\t\t%s" % self.args.user)
                print("Generated At:\t%s" % time.strftime(
                    "%Y-%m-%d %H:%M:%S UTC", time.gmtime(now)))
                print("Expired At:\t%s" % time.strftime(
                    "%Y-%m-%d %H:%M:%S UTC", time.gmtime(exp)))
                print("SHA1 Checksum:\t%s" % hashlib.sha1(
                    auth_token.encode("utf-8")).hexdigest())
                if self.args.claim:
                    print("Custom Claims:\t%s" % self.args.claim)
                print("-------------------------------------------")
            err_code = 0
        except Exception as e:
            print("Error when generating Auth Token")
            print(e)
            err_code = 1
        finally:
            sys.exit(err_code)

    def promote(self):
        client = self.get_client()
        r = client.promote(
            tenant=self.args.tenant,
            pipeline=self.args.pipeline,
            change_ids=self.args.changes)
        return r

    def show_running_jobs(self):
        client = self.get_client()
        running_items = client.get_running_jobs()

        if len(running_items) == 0:
            print("No jobs currently running")
            return True

        all_fields = self._show_running_jobs_columns()
        fields = all_fields.keys()

        table = prettytable.PrettyTable(
            field_names=[all_fields[f]['title'] for f in fields])
        for item in running_items:
            for job in item['jobs']:
                values = []
                for f in fields:
                    v = job
                    for part in f.split('.'):
                        if hasattr(v, 'get'):
                            v = v.get(part, '')
                    if ('transform' in all_fields[f]
                        and callable(all_fields[f]['transform'])):
                        v = all_fields[f]['transform'](v)
                    if 'append' in all_fields[f]:
                        v += all_fields[f]['append']
                    values.append(v)
                table.add_row(values)
        print(table)
        return True

    def _epoch_to_relative_time(self, epoch):
        if epoch:
            delta = datetime.timedelta(seconds=(time.time() - int(epoch)))
            return babel.dates.format_timedelta(delta, locale='en_US')
        else:
            return "Unknown"

    def _boolean_to_yes_no(self, value):
        return 'Yes' if value else 'No'

    def _boolean_to_pass_fail(self, value):
        return 'Pass' if value else 'Fail'

    def _format_list(self, l):
        return ', '.join(l) if isinstance(l, list) else ''

    def _show_running_jobs_columns(self):
        """A helper function to get the list of available columns for
        `zuul show running-jobs`. Also describes how to convert particular
        values (for example epoch to time string)"""

        return {
            'name': {
                'title': 'Job Name',
            },
            'elapsed_time': {
                'title': 'Elapsed Time',
                'transform': self._epoch_to_relative_time
            },
            'remaining_time': {
                'title': 'Remaining Time',
                'transform': self._epoch_to_relative_time
            },
            'url': {
                'title': 'URL'
            },
            'result': {
                'title': 'Result'
            },
            'voting': {
                'title': 'Voting',
                'transform': self._boolean_to_yes_no
            },
            'uuid': {
                'title': 'UUID'
            },
            'execute_time': {
                'title': 'Execute Time',
                'transform': self._epoch_to_relative_time,
                'append': ' ago'
            },
            'start_time': {
                'title': 'Start Time',
                'transform': self._epoch_to_relative_time,
                'append': ' ago'
            },
            'end_time': {
                'title': 'End Time',
                'transform': self._epoch_to_relative_time,
                'append': ' ago'
            },
            'estimated_time': {
                'title': 'Estimated Time',
                'transform': self._epoch_to_relative_time,
                'append': ' to go'
            },
            'pipeline': {
                'title': 'Pipeline'
            },
            'canceled': {
                'title': 'Canceled',
                'transform': self._boolean_to_yes_no
            },
            'retry': {
                'title': 'Retry'
            },
            'number': {
                'title': 'Number'
            },
            'nodeset': {
                'title': 'Nodeset'
            },
            'worker.name': {
                'title': 'Worker'
            },
            'worker.hostname': {
                'title': 'Worker Hostname'
            },
        }

    def validate(self):
        from zuul import scheduler
        from zuul import configloader
        from zuul.configloader import as_list
        self.configure_connections(sources=True, triggers=True, reporters=True)

        class SchedulerConfig(scheduler.Scheduler):
            # A custom scheduler constructor adapted for config check
            # to avoid loading runtime clients.
            def __init__(self, config, connections):
                self.config = config
                self.connections = connections
                self.unparsed_config_cache = None

        zuul_globals = SystemAttributes.fromConfig(self.config)
        sched = SchedulerConfig(self.config, self.connections)
        loader = configloader.ConfigLoader(
            self.connections, None, None, zuul_globals, None)
        tenant_config, script = sched._checkTenantSourceConf(self.config)
        try:
            abide = Abide()
            unparsed_abide = loader.readConfig(
                tenant_config, from_script=script)
            loader.loadAuthzRules(abide, unparsed_abide)
            loader.loadSemaphores(abide, unparsed_abide)
            # Do not load TPCs since that requires ZK
            for conf_tenant in unparsed_abide.tenants.values():
                loader.tenant_parser.getSchema()(conf_tenant)
                global_semaphores = set(as_list(conf_tenant.get(
                    'semaphores', [])))
                loader.tenant_parser.validateGlobalSemaphores(
                    abide, global_semaphores)

            print("Tenants config validated with success")
            err_code = 0
        except Exception as e:
            print("Error when validating tenants config")
            print(e)
            err_code = 1
        finally:
            sys.exit(err_code)

    def export_keys(self):
        logging.basicConfig(level=logging.INFO)

        zk_client = ZooKeeperClient.fromConfig(self.config)
        zk_client.connect()
        try:
            password = self.config["keystore"]["password"]
        except KeyError:
            raise RuntimeError("No key store password configured!")
        keystore = KeyStorage(zk_client, password=password)
        export = keystore.exportKeys()
        with open(os.open(self.args.path,
                          os.O_CREAT | os.O_WRONLY, 0o600), 'w') as f:
            json.dump(export, f)
        sys.exit(0)

    def import_keys(self):
        logging.basicConfig(level=logging.INFO)

        zk_client = ZooKeeperClient.fromConfig(self.config)
        zk_client.connect()
        try:
            password = self.config["keystore"]["password"]
        except KeyError:
            raise RuntimeError("No key store password configured!")
        keystore = KeyStorage(zk_client, password=password)
        with open(self.args.path, 'r') as f:
            import_data = json.load(f)
        keystore.importKeys(import_data, self.args.force)
        sys.exit(0)

    def copy_keys(self):
        logging.basicConfig(level=logging.INFO)

        zk_client = ZooKeeperClient.fromConfig(self.config)
        zk_client.connect()
        try:
            password = self.config["keystore"]["password"]
        except KeyError:
            raise RuntimeError("No key store password configured!")
        keystore = KeyStorage(zk_client, password=password)
        args = self.args
        # Load
        ssh = keystore.loadProjectSSHKeys(args.src_connection,
                                          args.src_project)
        secrets = keystore.loadProjectsSecretsKeys(args.src_connection,
                                                   args.src_project)
        # Save
        keystore.saveProjectSSHKeys(args.dest_connection,
                                    args.dest_project, ssh)
        keystore.saveProjectsSecretsKeys(args.dest_connection,
                                         args.dest_project, secrets)
        self.log.info("Copied keys from %s %s to %s %s",
                      args.src_connection, args.src_project,
                      args.dest_connection, args.dest_project)
        sys.exit(0)

    def delete_keys(self):
        logging.basicConfig(level=logging.INFO)

        zk_client = ZooKeeperClient.fromConfig(self.config)
        zk_client.connect()
        try:
            password = self.config["keystore"]["password"]
        except KeyError:
            raise RuntimeError("No key store password configured!")
        keystore = KeyStorage(zk_client, password=password)
        args = self.args
        keystore.deleteProjectSSHKeys(args.connection, args.project)
        keystore.deleteProjectsSecretsKeys(args.connection, args.project)
        keystore.deleteProjectDir(args.connection, args.project)
        self.log.info("Delete keys from %s %s",
                      args.connection, args.project)
        sys.exit(0)

    def delete_oidc_signing_keys(self):
        logging.basicConfig(level=logging.INFO)

        zk_client = ZooKeeperClient.fromConfig(self.config)
        zk_client.connect()
        try:
            password = self.config["keystore"]["password"]
        except KeyError:
            raise RuntimeError("No key store password configured!")
        keystore = KeyStorage(zk_client, password=password)
        algorithm = self.args.algorithm
        keystore.deleteOidcSigningKeys(algorithm)
        self.log.info("Delete OIDC signing keys for %s", algorithm)
        sys.exit(0)

    def export_images(self):
        logging.basicConfig(level=logging.INFO)

        zk_client = ZooKeeperClient.fromConfig(self.config)
        zk_client.connect()
        image_build_registry = ImageBuildRegistry(zk_client)
        image_build_registry.waitForSync()
        image_upload_registry = ImageUploadRegistry(zk_client)
        image_upload_registry.waitForSync()
        artifacts = [
            x.toDict() for x in image_build_registry.getAllArtifacts()]
        uploads = [
            x.toDict() for x in image_upload_registry.getAllUploads()]
        export = dict(
            artifacts=artifacts,
            uploads=uploads,
        )

        with open(os.open(self.args.path,
                          os.O_CREAT | os.O_WRONLY, 0o600), 'w') as f:
            json.dump(export, f)
        sys.exit(0)

    def import_images(self):
        logging.basicConfig(level=logging.INFO)

        zk_client = ZooKeeperClient.fromConfig(self.config)
        zk_client.connect()
        with open(self.args.path, 'r') as f:
            import_data = json.load(f)
        with ZKContext(zk_client, None, None, self.log) as context:
            for upload in import_data['uploads']:
                try:
                    ImageUpload.new(context, **upload)
                except NodeExistsError:
                    self.log.debug("Upload already exists %s", upload['uuid'])
            for artifact in import_data['artifacts']:
                try:
                    ImageBuildArtifact.new(context, **artifact)
                except NodeExistsError:
                    self.log.debug("Artifact already exists %s",
                                   artifact['uuid'])

        sys.exit(0)

    def delete_state(self):
        logging.basicConfig(level=logging.INFO)

        zk_client = ZooKeeperClient.fromConfig(self.config)
        zk_client.connect()
        confirm = input("Are you sure you want to delete "
                        "all ephemeral data from ZooKeeper? (yes/no) ")
        if confirm.strip().lower() != 'yes':
            print("Aborting")
            sys.exit(1)
        if self.args.keep_config_cache:
            try:
                children = zk_client.client.get_children('/zuul')
            except NoNodeError:
                children = []
            for child in children:
                if child == 'config':
                    continue
                path = f'/zuul/{child}'
                self.log.debug("Deleting %s", path)
                zk_client.fastRecursiveDelete(path)
        else:
            zk_client.fastRecursiveDelete('/zuul')
        sys.exit(0)

    def delete_pipeline_state(self):
        logging.basicConfig(level=logging.INFO)

        zk_client = ZooKeeperClient.fromConfig(self.config)
        zk_client.connect()

        args = self.args
        safe_tenant = urllib.parse.quote_plus(args.tenant)
        safe_pipeline = urllib.parse.quote_plus(args.pipeline)
        COMPONENT_REGISTRY.create(zk_client)

        class FakeTenant:
            pass

        tenant = FakeTenant()
        tenant.name = args.tenant

        with tenant_read_lock(zk_client, args.tenant, self.log):
            path = f'/zuul/tenant/{safe_tenant}/pipeline/{safe_pipeline}'
            pipeline = Pipeline(args.pipeline)
            with pipeline_lock(
                    zk_client, args.tenant, args.pipeline
            ) as plock:
                zk_client.fastRecursiveDelete(path)
                with ZKContext(zk_client, plock, None, self.log) as context:
                    manager = IndependentPipelineManager(
                        None, pipeline, tenant)
                    manager.state = PipelineState.new(
                        context, _path=path, layout_uuid=None)
                    PipelineChangeList.new(context, manager=manager)

        sys.exit(0)

    def prune_database(self):
        logging.basicConfig(level=logging.INFO)
        args = self.args
        now = datetime.datetime.now(dateutil.tz.tzutc())
        cutoff = parse_cutoff(now, args.before, args.older_than)
        self.configure_connections(database=True)
        connection = self.connections.getSqlConnection()
        connection.deleteBuildsets(cutoff, args.batch_size)
        connection.deleteSystemEvents(cutoff)
        sys.exit(0)


def main():
    if sys.argv[0].endswith('zuul'):
        print(
            "Warning: this command name is deprecated, "
            "use `zuul-admin` instead",
            file=sys.stderr)
    Client().main()



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import yaml
import logging

from datetime import datetime
from elasticsearch import Elasticsearch
from elasticsearch.client import IndicesClient
from elasticsearch.helpers import bulk
from elasticsearch.helpers import BulkIndexError

from zuul.connection import BaseConnection


class ElasticsearchConnection(BaseConnection):
    driver_name = 'elasticsearch'
    log = logging.getLogger("zuul.ElasticSearchConnection")
    properties = {
        #  Common attribute
        "uuid": {"type": "keyword"},
        "build_type": {"type": "keyword"},
        "result": {"type": "keyword"},
        "duration": {"type": "integer"},
        "@timestamp": {"type": "date"},
        # BuildSet type specific attributes
        "zuul_ref": {"type": "keyword"},
        "pipeline": {"type": "keyword"},
        "project": {"type": "keyword"},
        "branch": {"type": "keyword"},
        "change": {"type": "integer"},
        "patchset": {"type": "keyword"},
        "ref": {"type": "keyword"},
        "oldrev": {"type": "keyword"},
        "newrev": {"type": "keyword"},
        "ref_url": {"type": "keyword"},
        "message": {"type": "text"},
        "tenant": {"type": "keyword"},
        # Build type specific attibutes
        "buildset_uuid": {"type": "keyword"},
        "job_name": {"type": "keyword"},
        "start_time": {"type": "date", "format": "epoch_second"},
        "end_time": {"type": "date", "format": "epoch_second"},
        "voting": {"type": "boolean"},
        "log_url": {"type": "keyword"},
        "nodeset": {"type": "keyword"}
    }

    def __init__(self, driver, connection_name, connection_config):
        super(ElasticsearchConnection, self).__init__(
            driver, connection_name, connection_config)
        self.uri = self.connection_config.get('uri').split(',')
        self.cnx_opts = {}
        use_ssl = self.connection_config.get('use_ssl', True)
        if isinstance(use_ssl, str):
            if use_ssl.lower() == 'false':
                use_ssl = False
            else:
                use_ssl = True
        self.cnx_opts['use_ssl'] = use_ssl
        if use_ssl:
            verify_certs = self.connection_config.get('verify_certs', True)
            if isinstance(verify_certs, str):
                if verify_certs.lower() == 'false':
                    verify_certs = False
                else:
                    verify_certs = True
            self.cnx_opts['verify_certs'] = verify_certs
            self.cnx_opts['ca_certs'] = self.connection_config.get(
                'ca_certs', None)
            self.cnx_opts['client_cert'] = self.connection_config.get(
                'client_cert', None)
            self.cnx_opts['client_key'] = self.connection_config.get(
                'client_key', None)
        self.es = Elasticsearch(
            self.uri, **self.cnx_opts)
        try:
            self.log.debug("Elasticsearch info: %s" % self.es.info())
        except Exception as e:
            self.log.warn("An error occured on estabilishing "
                          "connection to Elasticsearch: %s" % e)
        self.ic = IndicesClient(self.es)

    def setIndex(self, index):
        settings = {
            'mappings': {
                '_doc': {
                    "properties": self.properties
                }
            }
        }
        try:
            self.ic.create(index=index, ignore=400, body=settings)
        except Exception:
            self.log.exception(
                "Unable to create the index %s on connection %s" % (
                    index, self.connection_name))

    def gen(self, it, index):
        for source in it:
            d = {}
            source['@timestamp'] = datetime.utcfromtimestamp(
                int(source['start_time'])).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            d['_index'] = index
            d['_op_type'] = 'index'
            d['_source'] = source
            yield d

    def add_docs(self, source_it, index):

        self.setIndex(index)

        try:
            bulk(self.es, self.gen(source_it, index))
            self.es.indices.refresh(index=index)
            self.log.debug('%s docs indexed to %s' % (
                len(source_it), self.connection_name))
        except BulkIndexError as exc:
            self.log.warn("Some docs failed to be indexed (%s)" % exc)
            # We give flexibility by allowing any type of job's vars and
            # zuul return data to be indexed with EL dynamic mapping enabled.
            # It may happen that a doc own a field with a value that does not
            # match the previous data type that EL has detected for that field.
            # In that case the whole doc is not indexed by EL.
            # Here we want to mitigate by indexing the errorneous docs in a
            # <index-name>.errorneous index by flattening the doc data as yaml.
            # This ensures the doc is indexed and can be tracked and eventually
            # be modified and re-indexed by an operator.
            errorneous_docs = []
            for d in exc.errors:
                if d['index']['error']['type'] == 'mapper_parsing_exception':
                    errorneous_doc = {
                        'uuid': d['index']['data']['uuid'],
                        'blob': yaml.dump(d['index']['data'])
                    }
                    errorneous_docs.append(errorneous_doc)
            try:
                mapping_errorneous_index = "%s.errorneous" % index
                bulk(
                    self.es,
                    self.gen(errorneous_docs, mapping_errorneous_index))
                self.es.indices.refresh(index=mapping_errorneous_index)
                self.log.info(
                    "%s errorneous docs indexed" % (len(errorneous_docs)))
            except BulkIndexError as exc:
                self.log.warn(
                    "Some errorneous docs failed to be indexed (%s)" % exc)



# Copyright 2015 Rackspace Australia
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import json
import logging
import time
import voluptuous as v

import sqlalchemy.exc

from zuul.driver.sql.sqlconnection import SQL_MAX_STRING_LENGTH
from zuul.exceptions import MissingBuildsetError
from zuul.lib.result_data import get_artifacts_from_result_data
from zuul.reporter import BaseReporter


DATA_LENGTH_ERROR = """
%s's length exceeds DB storage capacity.
Original value: %s"""


class SQLReporter(BaseReporter):
    """Sends off reports to a database."""

    name = 'sql'
    log = logging.getLogger("zuul.SQLReporter")
    retry_count = 3
    retry_delay = 5

    def _getBuildData(self, item, job, build):
        (result, _) = item.formatJobResult(job, build)
        start = end = None
        if build.start_time:
            start = datetime.datetime.fromtimestamp(
                build.start_time,
                tz=datetime.timezone.utc)
        if build.end_time:
            end = datetime.datetime.fromtimestamp(
                build.end_time,
                tz=datetime.timezone.utc)
        return result, build.log_url, start, end

    def _createBuildset(self, db, buildset):
        event_id = None
        event_timestamp = None
        item = buildset.item
        if item.event is not None:
            event_id = getattr(item.event, "zuul_event_id", None)
            event_timestamp = datetime.datetime.fromtimestamp(
                item.event.timestamp, tz=datetime.timezone.utc)

        db_buildset = db.createBuildSet(
            uuid=buildset.uuid,
            tenant=item.manager.tenant.name,
            pipeline=item.manager.pipeline.name,
            event_id=event_id,
            event_timestamp=event_timestamp,
            queue_item_uuid=item.uuid,
            updated=datetime.datetime.utcnow(),
        )
        for change in item.changes:
            ref = db.getOrCreateRef(
                project=change.project.name,
                change=getattr(change, 'number', None),
                patchset=getattr(change, 'patchset', None),
                ref_url=change.url,
                ref=getattr(change, 'ref', ''),
                oldrev=getattr(change, 'oldrev', ''),
                newrev=getattr(change, 'newrev', ''),
                branch=getattr(change, 'branch', ''),
            )
            db_buildset.refs.append(ref)
        event_change = item.getEventChange()
        if event_change:
            db_buildset.createBuildSetEvent(
                event_time=datetime.datetime.fromtimestamp(
                    item.event.timestamp, tz=datetime.timezone.utc),
                event_type='triggered',
                description=f'Triggered by {event_change.toString()}',
            )
        return db_buildset

    def reportBuildsetStart(self, buildset):
        """Create the initial buildset entry in the db"""
        if not buildset.uuid:
            return

        for retry_count in range(self.retry_count):
            try:
                with self.connection.getSession() as db:
                    return self._createBuildset(db, buildset)
            except sqlalchemy.exc.DBAPIError:
                if retry_count < self.retry_count - 1:
                    self.log.error("Unable to create buildset, will retry")
                    time.sleep(self.retry_delay)
                else:
                    self.log.exception("Unable to create buildset")

    def reportBuildsetEnd(self, buildset, action, final, result=None):
        if not buildset.uuid:
            return
        if final:
            message = self._formatItemReport(
                buildset.item, with_jobs=False, action=action)
        else:
            message = None
        for retry_count in range(self.retry_count):
            try:
                with self.connection.getSession() as db:
                    db_buildset = db.getBuildset(
                        tenant=buildset.item.manager.tenant.name,
                        uuid=buildset.uuid)
                    if not db_buildset:
                        db_buildset = self._createBuildset(db, buildset)
                    db_buildset.result = buildset.result or result
                    db_buildset.message = message
                    end_time = db_buildset.first_build_start_time
                    for build in db_buildset.builds:
                        if (build.end_time and end_time
                            and build.end_time > end_time):
                            end_time = build.end_time
                    db_buildset.last_build_end_time = end_time
                    db_buildset.updated = datetime.datetime.utcnow()
                    return
            except sqlalchemy.exc.DBAPIError:
                if retry_count < self.retry_count - 1:
                    self.log.error("Unable to update buildset, will retry")
                    time.sleep(self.retry_delay)
                else:
                    self.log.exception("Unable to update buildset")

    def reportBuildStart(self, build):
        for retry_count in range(self.retry_count):
            try:
                with self.connection.getSession() as db:
                    db_build = self._createBuild(db, build)
                    return db_build
            except sqlalchemy.exc.DBAPIError:
                if retry_count < self.retry_count - 1:
                    self.log.error("Unable to create build, will retry")
                    time.sleep(self.retry_delay)
                else:
                    self.log.exception("Unable to create build")

    def reportBuildEnd(self, build, tenant, final):
        return self.reportBuildEnds([build], tenant, final)

    def reportBuildEnds(self, builds, tenant, final):
        for retry_count in range(self.retry_count):
            try:
                with self.connection.getSession() as db:
                    buildset = builds[0].build_set
                    try:
                        db_buildset = self._getBuildset(db, builds[0])
                    except MissingBuildsetError:
                        # let _createBuild handle if necessary
                        pass
                    for build in builds:
                        if build.build_set is not buildset:
                            raise Exception("All batch reported builds "
                                            "must be for the same buildset")
                        self._reportBuildEnd(
                            db, db_buildset, build, tenant, final)
                # Exit retry loop
                return
            except sqlalchemy.exc.DBAPIError:
                if retry_count < self.retry_count - 1:
                    self.log.error("Unable to update build, will retry")
                    time.sleep(self.retry_delay)
                else:
                    self.log.exception("Unable to update build")

    def _reportBuildEnd(self, db, db_buildset, build, tenant, final):
        db_build = db.getBuild(tenant=tenant, uuid=build.uuid)
        if not db_build:
            db_build = self._createBuild(db, build, db_buildset=db_buildset)
        end_time = build.end_time or time.time()
        end = datetime.datetime.fromtimestamp(
            end_time, tz=datetime.timezone.utc)

        db_build.result = build.result
        db_build.end_time = end
        db_build.error_detail = build.error_detail
        db_build.log_url = build.log_url
        if build.log_url is not None:
            if len(build.log_url) >= SQL_MAX_STRING_LENGTH:
                db_build.log_url = None
                if db_build.error_detail is not None:
                    db_build.error_detail += (
                        DATA_LENGTH_ERROR % ("log URL", build.log_url)
                    )
                else:
                    db_build.error_detail = (
                        DATA_LENGTH_ERROR % ("log URL", build.log_url)
                    )
        db_build.final = final
        db_build.held = build.held

        for provides in build.job.provides:
            db_build.createProvides(name=provides)

        for artifact in get_artifacts_from_result_data(
                build.result_data,
                logger=self.log):
            if 'metadata' in artifact:
                artifact['metadata'] = json.dumps(
                    artifact['metadata'])
            if artifact['name'] is not None:
                if len(artifact['name']) > SQL_MAX_STRING_LENGTH:
                    db_build.error_detail = (db_build.error_detail or "") + (
                        DATA_LENGTH_ERROR % ("artifact name", artifact['name'])
                    )

                    artifact['name'] = (
                        artifact['name'][:SQL_MAX_STRING_LENGTH - 15] +
                        '... (truncated)'
                    )
            db_build.createArtifact(**artifact)

        for event in build.events:
            # Reformat the event_time so it's compatible to SQL.
            # Don't update the event object in place, but only
            # the generated dict representation to not alter the
            # datastructure for other reporters.
            ev = event.toDict()
            ev["event_time"] = datetime.datetime.fromtimestamp(
                event.event_time, tz=datetime.timezone.utc)
            db_build.createBuildEvent(**ev)
        return db_build

    def _getBuildset(self, db, build):
        buildset = build.build_set
        if not buildset:
            return None
        db_buildset = db.getBuildset(
            tenant=buildset.item.manager.tenant.name, uuid=buildset.uuid)
        if not db_buildset:
            self.log.warning("Creating missing buildset %s", buildset.uuid)
            db_buildset = self._createBuildset(db, buildset)
        return db_buildset

    def _createBuild(self, db, build, db_buildset=None):
        start_time = build.start_time or time.time()
        start = datetime.datetime.fromtimestamp(start_time,
                                                tz=datetime.timezone.utc)
        execute_time = datetime.datetime.fromtimestamp(
            build.execute_time or time.time(), tz=datetime.timezone.utc)
        if db_buildset is None:
            db_buildset = self._getBuildset(db, build)
        if db_buildset is None:
            raise MissingBuildsetError()
        if db_buildset.first_build_start_time is None:
            db_buildset.first_build_start_time = start
        item = build.build_set.item
        change = item.getChangeForJob(build.job)
        ref = db.getOrCreateRef(
            project=change.project.name,
            change=getattr(change, 'number', None),
            patchset=getattr(change, 'patchset', None),
            ref_url=change.url,
            ref=getattr(change, 'ref', ''),
            oldrev=getattr(change, 'oldrev', ''),
            newrev=getattr(change, 'newrev', ''),
            branch=getattr(change, 'branch', ''),
        )

        db_build = db_buildset.createBuild(
            ref=ref,
            uuid=build.uuid,
            job_name=build.job.name,
            execute_time=execute_time,
            start_time=start,
            voting=build.job.voting,
            nodeset=build.job.nodeset.name,
        )
        return db_build

    def getBuilds(self, *args, **kw):
        """Return a list of Build objects"""
        return self.connection.getBuilds(*args, **kw)

    def reportSystemEvent(self, tenant, **kw):
        if 'event_time' in kw:
            kw['event_time'] = datetime.datetime.fromtimestamp(
                kw['event_time'], tz=datetime.timezone.utc)

        for retry_count in range(self.retry_count):
            try:
                with self.connection.getSession() as db:
                    db_event = db.createSystemEvent(tenant=tenant, **kw)
                    return db_event
            except sqlalchemy.exc.DBAPIError:
                if retry_count < self.retry_count - 1:
                    self.log.error("Unable to create event, will retry")
                    time.sleep(self.retry_delay)
                else:
                    self.log.exception("Unable to create event")

    def report(self, item):
        # We're not a real reporter, but we use _formatItemReport, so
        # we inherit from the reporters.
        raise NotImplementedError()


def getSchema():
    sql_reporter = v.Schema(None)
    return sql_reporter



# Copyright 2015 Christoph Gysin <christoph.gysin@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import requests

from urllib.parse import urlparse, urlunparse


class FormAuth(requests.auth.AuthBase):
    log = logging.getLogger('zuul.GerritConnection')

    def __init__(self, username, password):
        self.username = username
        self.password = password

    def _retry_using_form_auth(self, response, args):
        adapter = requests.adapters.HTTPAdapter()
        request = _copy_request(response.request)

        u = urlparse(response.url)
        url = urlunparse([u.scheme, u.netloc, '/login', None, None, None])
        auth = {'username': self.username,
                'password': self.password}
        request2 = requests.Request('POST', url, data=auth).prepare()
        response2 = adapter.send(request2, **args)

        if response2.status_code == 401:
            self.log.error('Login failed: Invalid username or password?')
            return response

        cookie = response2.headers.get('set-cookie')
        if cookie is not None:
            request.headers['Cookie'] = cookie

        response3 = adapter.send(request, **args)
        return response3

    def _response_hook(self, response, **kwargs):
        if response.status_code == 401:
            return self._retry_using_form_auth(response, kwargs)
        return response

    def __call__(self, request):
        request.headers["Connection"] = "Keep-Alive"
        request.register_hook('response', self._response_hook)
        return request


def _copy_request(request):
    new_request = requests.PreparedRequest()
    new_request.method = request.method
    new_request.url = request.url
    new_request.body = request.body
    new_request.hooks = request.hooks
    new_request.headers = request.headers.copy()
    return new_request



# Copyright 2021 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import configparser

from zuul.lib.fingergw import FingerGateway
from zuul.zk.components import BaseComponent, ComponentRegistry

from tests.base import (
    iterate_timeout,
    okay_tracebacks,
    ZuulTestCase,
    ZuulWebFixture,
)


class TestComponentRegistry(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def setUp(self):
        super().setUp()

        self.component_registry = ComponentRegistry(self.zk_client)

    def assertComponentAttr(self, component_name, attr_name,
                            attr_value, timeout=25):
        for _ in iterate_timeout(
            timeout,
            f"{component_name} in cache has {attr_name} set to {attr_value}",
        ):
            components = list(self.component_registry.all(component_name))
            if (
                len(components) > 0 and
                getattr(components[0], attr_name) == attr_value
            ):
                break

    def assertComponentState(self, component_name, state, timeout=25):
        return self.assertComponentAttr(
            component_name, "state", state, timeout
        )

    def assertComponentStopped(self, component_name, timeout=25):
        for _ in iterate_timeout(
            timeout, f"{component_name} in cache is stopped"
        ):
            components = list(self.component_registry.all(component_name))
            if len(components) == 0:
                break

    def test_scheduler_component(self):
        self.assertComponentState("scheduler", BaseComponent.RUNNING)

    @okay_tracebacks('_start',
                     '_playbackWorker')
    def test_executor_component(self):
        self.assertComponentState("executor", BaseComponent.RUNNING)

        self.executor_server.pause()
        self.assertComponentState("executor", BaseComponent.PAUSED)

        self.executor_server.unpause()
        self.assertComponentState("executor", BaseComponent.RUNNING)

        self.executor_server.unregister_work()
        self.assertComponentAttr("executor", "accepting_work", False)

        self.executor_server.register_work()
        self.assertComponentAttr("executor", "accepting_work", True)

        # This can cause tracebacks in the logs when the tree cache
        # attempts to restart.
        self.executor_server.zk_client.client.stop()
        self.assertComponentStopped("executor")

        self.executor_server.zk_client.client.start()
        self.assertComponentAttr("executor", "accepting_work", True)

    def test_merger_component(self):
        self._startMerger()
        self.assertComponentState("merger", BaseComponent.RUNNING)

        self.merge_server.pause()
        self.assertComponentState("merger", BaseComponent.PAUSED)

        self.merge_server.unpause()
        self.assertComponentState("merger", BaseComponent.RUNNING)

        self.merge_server.stop()
        self.merge_server.join()
        # Set the merger to None so the test doesn't try to stop it again
        self.merge_server = None

        try:
            self.assertComponentStopped("merger")
        except Exception:
            for kind, components in self.component_registry.all():
                self.log.error("Component %s has %s online", kind, components)
            raise

    def test_fingergw_component(self):
        config = configparser.ConfigParser()
        config.read_dict(self.config)
        config.read_dict({
            'fingergw': {
                'listen_address': '::',
                'port': '0',
                'hostname': 'janine',
            }
        })
        gateway = FingerGateway(
            config,
            command_socket=None,
            pid_file=None
        )
        gateway.start()

        try:
            self.assertComponentState("fingergw", BaseComponent.RUNNING)
            self.assertComponentAttr("fingergw", "hostname", "janine")
        finally:
            gateway.stop()

        self.assertComponentStopped("fingergw")

    def test_web_component(self):
        self.useFixture(
            ZuulWebFixture(
                self.config, self.test_config, self.additional_event_queues,
                self.upstream_root, self.poller_events,
                self.git_url_with_auth, self.addCleanup, self.test_root
            )
        )

        self.assertComponentState("web", BaseComponent.RUNNING)

    def test_launcher_component(self):
        self.assertComponentState("launcher", BaseComponent.RUNNING)



# Copyright 2018 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Hook for pbr to build javascript as part of tarball."""

import os
import subprocess


def _build_javascript():
    if subprocess.call(['which', 'yarn']) != 0:
        return
    if not os.path.exists('web/node_modules/.bin/webpack'):
        r = subprocess.Popen(['yarn', 'install',
                              '--frozen-lockfile'], cwd="web/").wait()
        if r:
            raise RuntimeError("Yarn install failed")
    r = subprocess.Popen(['yarn', 'list'], cwd="web/").wait()
    if r:
        raise RuntimeError("Yarn list failed")
    if not os.path.exists('zuul/web/static/index.html'):
        os.makedirs('zuul/web/static', exist_ok=True)
        if not os.path.islink('../zuul/web/static'):
            os.symlink('../zuul/web/static', 'web/dist',
                       target_is_directory=True)
        r = subprocess.Popen(['yarn', 'build'], cwd="web/").wait()
        if r:
            raise RuntimeError("Yarn build failed")


def setup_hook(config):
    _build_javascript()



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tests.base import ZuulTestCase


class TestUpgradeOld(ZuulTestCase):
    tenant_config_file = "config/single-tenant/main.yaml"
    scheduler_count = 1
    random_databases = False
    delete_databases = False
    use_tmpdir = False
    init_repos = True
    load_change_db = False
    always_attach_logs = True

    def assertFinalState(self):
        # In this test, we expect to shut down in a non-final state,
        # so skip these checks.
        pass

    def shutdown(self):
        # Shutdown the scheduler now before it gets any aborted events
        if self.validate_tenants is None:
            self.scheds.execute(lambda app: app.sched.stop())
            self.scheds.execute(lambda app: app.sched.join())
        else:
            self.scheds.execute(lambda app: app.sched.stopConnections())
        # Then release the executor jobs and stop the executors
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.scheds.execute(lambda app: app.sched.executor.stop())
        if self.merge_server:
            self.merge_server.stop()
            self.merge_server.join()

        self.executor_server.stop()
        self.executor_server.join()
        self.launcher.stop()
        self.launcher.join()
        self.statsd.stop()
        self.statsd.join()
        self.fake_nodepool.stop()
        self.zk_client.disconnect()
        self.printHistory()

    def test_upgrade(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertHistory([])
        self.assertEqual(len(self.builds), 1)
        self.executor_server.release()
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project-merge', result='SUCCESS', changes='1,1'),
        ], ordered=False)
        self.assertEqual(len(self.builds), 2)
        self.saveChangeDB()



# Copyright 2015 Red Hat, Inc.
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import configparser
import os
import textwrap
import uuid
from unittest import mock

import fixtures
import testtools
import voluptuous

from zuul import model
from zuul import configloader
from zuul.lib import encryption
from zuul.lib import yamlutil as yaml
import zuul.lib.connections

from tests.base import BaseTestCase, FIXTURE_DIR
from zuul.lib.ansible import AnsibleManager
from zuul.lib import tracing
from zuul.lib.re2util import ZuulRegex
from zuul.model_api import MODEL_API
from zuul.zk.zkobject import LocalZKContext
from zuul.zk.components import COMPONENT_REGISTRY
from zuul import change_matcher


class Dummy(object):
    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)


class TestJob(BaseTestCase):
    def setUp(self):
        COMPONENT_REGISTRY.registry = Dummy()
        COMPONENT_REGISTRY.registry.model_api = MODEL_API
        self._env_fixture = self.useFixture(
            fixtures.EnvironmentVariable('HISTTIMEFORMAT', '%Y-%m-%dT%T%z '))
        super(TestJob, self).setUp()
        # Toss in % in env vars to trigger the configparser issue
        self.connections = zuul.lib.connections.ConnectionRegistry()
        self.addCleanup(self.connections.stop)
        self.connection = Dummy(connection_name='dummy_connection')
        self.source = Dummy(canonical_hostname='git.example.com',
                            connection=self.connection)
        self.abide = model.Abide()
        self.tenant = model.Tenant('tenant')
        self.tenant.allowed_labels = None
        self.tenant.disallowed_labels = None
        self.tenant.default_ansible_version = AnsibleManager().default_version
        self.tenant.semaphore_handler = Dummy(abide=self.abide)
        self.layout = model.Layout(self.tenant)
        self.tenant.layout = self.layout
        self.project = model.Project('project', self.source)
        self.context = model.SourceContext(
            self.project.canonical_name, self.project.name,
            self.project.connection_name, 'master', 'test')
        self.tpc = model.TenantProjectConfig(self.project)
        self.tpc.trusted = True
        self.tenant.addTPC(self.tpc)
        self.pipeline = model.Pipeline('gate')
        self.pipeline.source_context = self.context
        self.manager = mock.Mock()
        self.manager.pipeline = self.pipeline
        self.manager.tenant = self.tenant
        self.zk_context = LocalZKContext(self.log)
        self.manager.current_context = self.zk_context
        self.manager.state = model.PipelineState()
        self.manager.state._set(manager=self.manager)
        self.layout.addPipelineManager(self.manager)
        with self.zk_context as ctx:
            self.queue = model.ChangeQueue.new(
                ctx, manager=self.manager)
        self.pcontext = configloader.ParseContext(
            self.connections, None, None, AnsibleManager())

        private_key_file = os.path.join(FIXTURE_DIR, 'private.pem')
        with open(private_key_file, "rb") as f:
            priv, pub = encryption.deserialize_rsa_keypair(f.read())
            self.project.private_secrets_key = priv
            self.project.public_secrets_key = pub
        m = yaml.Mark('name', 0, 0, 0, '', 0)
        self.start_mark = model.ZuulMark(m, m, '')
        config = configparser.ConfigParser()
        self.tracing = tracing.Tracing(config)

    @property
    def job(self):
        job = self.pcontext.job_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'job',
            'parent': None,
            'irrelevant-files': [
                '^docs/.*$'
            ]}, None)
        return job

    def test_change_matches_returns_false_for_matched_skip_if(self):
        change = model.Change('project')
        change.files = ['/COMMIT_MSG', 'docs/foo']
        self.assertFalse(self.job.changeMatchesFiles(change))

    def test_change_matches_returns_false_for_single_matched_skip_if(self):
        change = model.Change('project')
        change.files = ['docs/foo']
        self.assertFalse(self.job.changeMatchesFiles(change))

    def test_change_matches_returns_true_for_unmatched_skip_if(self):
        change = model.Change('project')
        change.files = ['/COMMIT_MSG', 'foo']
        self.assertTrue(self.job.changeMatchesFiles(change))

    def test_change_matches_returns_true_for_single_unmatched_skip_if(self):
        change = model.Change('project')
        change.files = ['foo']
        self.assertTrue(self.job.changeMatchesFiles(change))

    def test_job_sets_defaults_for_boolean_attributes(self):
        self.assertIsNotNone(self.job.voting)

    def test_job_variants(self):
        # This simulates freezing a job.

        secret = model.Secret('foo', self.context)
        self.layout.addSecret(secret)

        secrets = [model.SecretUse('foo', 'foo')]
        py27_pre = model.PlaybookContext(
            self.context, 'py27-pre', [], secrets, [])
        py27_run = model.PlaybookContext(
            self.context, 'py27-run', [], secrets, [])
        py27_post = model.PlaybookContext(
            self.context, 'py27-post', [], secrets, [])

        py27 = model.Job('py27')
        py27.timeout = 30
        py27.pre_run = (py27_pre,)
        py27.run = (py27_run,)
        py27.post_run = (py27_post,)

        job = py27.copy()
        job.setBase(self.layout, None)
        self.assertEqual(30, job.timeout)

        # Apply the diablo variant
        diablo = model.Job('py27')
        diablo.timeout = 40
        job.applyVariant(diablo, self.layout, None)

        self.assertEqual(40, job.timeout)
        self.assertEqual(['py27-pre'],
                         [x.path for x in job.pre_run])
        self.assertEqual(['py27-run'],
                         [x.path for x in job.run])
        self.assertEqual(['py27-post'],
                         [x.path for x in job.post_run])
        self.assertEqual(secrets, job.pre_run[0].secrets)
        self.assertEqual(secrets, job.run[0].secrets)
        self.assertEqual(secrets, job.post_run[0].secrets)

        # Set the job to final for the following checks
        job.final = True
        self.assertTrue(job.voting)

        good_final = model.Job('py27')
        good_final.voting = False
        job.applyVariant(good_final, self.layout, None)
        self.assertFalse(job.voting)

        bad_final = model.Job('py27')
        bad_final.timeout = 600
        with testtools.ExpectedException(
                model.JobConfigurationError,
                "Unable to modify final job"):
            job.applyVariant(bad_final, self.layout, None)

    @mock.patch("zuul.model.zkobject.ZKObject._save")
    def test_job_inheritance_job_tree(self, save_mock):
        base = self.pcontext.job_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'base',
            'parent': None,
            'timeout': 30,
        }, None)
        self.layout.addJob(base)
        python27 = self.pcontext.job_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'python27',
            'parent': 'base',
            'timeout': 40,
        }, None)
        self.layout.addJob(python27)
        python27diablo = self.pcontext.job_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'python27',
            'branches': [
                'stable/diablo'
            ],
            'timeout': 50,
        }, None)
        self.layout.addJob(python27diablo)

        project_config = self.pcontext.project_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'project',
            'gate': {
                'jobs': [
                    {'python27': {'timeout': 70,
                                  'run': 'playbooks/python27.yaml'}}
                ]
            }
        })
        self.layout.addProjectConfig(self.project.canonical_name,
                                     project_config)

        change = model.Change(self.project)
        change.branch = 'master'
        change.cache_stat = Dummy(key=Dummy(reference=uuid.uuid4().hex))
        item = self.queue.enqueueChanges([change], None)

        self.assertTrue(base.changeMatchesBranch(self.tenant, change))
        self.assertTrue(python27.changeMatchesBranch(self.tenant, change))
        self.assertFalse(python27diablo.changeMatchesBranch(
            self.tenant, change))

        with self.zk_context as ctx:
            item.freezeJobGraph(self.layout, ctx,
                                skip_file_matcher=False,
                                redact_secrets_and_keys=False)
        self.assertEqual(len(item.getJobs()), 1)
        job = item.getJobs()[0]
        self.assertEqual(job.name, 'python27')
        self.assertEqual(job.timeout, 70)

        change.branch = 'stable/diablo'
        change.cache_stat = Dummy(key=Dummy(reference=uuid.uuid4().hex))
        item = self.queue.enqueueChanges([change], None)

        self.assertTrue(base.changeMatchesBranch(self.tenant, change))
        self.assertTrue(python27.changeMatchesBranch(self.tenant, change))
        self.assertTrue(python27diablo.changeMatchesBranch(
            self.tenant, change))

        with self.zk_context as ctx:
            item.freezeJobGraph(self.layout, ctx,
                                skip_file_matcher=False,
                                redact_secrets_and_keys=False)
        self.assertEqual(len(item.getJobs()), 1)
        job = item.getJobs()[0]
        self.assertEqual(job.name, 'python27')
        self.assertEqual(job.timeout, 70)

    @mock.patch("zuul.model.zkobject.ZKObject._save")
    def test_inheritance_keeps_matchers(self, save_mock):
        base = self.pcontext.job_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'base',
            'parent': None,
            'timeout': 30,
        }, None)
        self.layout.addJob(base)
        python27 = self.pcontext.job_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'python27',
            'parent': 'base',
            'timeout': 40,
            'irrelevant-files': ['^ignored-file$'],
        }, None)
        self.layout.addJob(python27)

        project_config = self.pcontext.project_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'project',
            'gate': {
                'jobs': [
                    'python27',
                ]
            }
        })
        self.layout.addProjectConfig(self.project.canonical_name,
                                     project_config)

        change = model.Change(self.project)
        change.branch = 'master'
        change.cache_stat = Dummy(key=Dummy(reference=uuid.uuid4().hex))
        change.files = ['/COMMIT_MSG', 'ignored-file']
        item = self.queue.enqueueChanges([change], None)

        self.assertTrue(base.changeMatchesFiles(change))
        self.assertFalse(python27.changeMatchesFiles(change))

        self.manager.getFallbackLayout = mock.Mock(return_value=None)
        with self.zk_context as ctx:
            item.freezeJobGraph(self.layout, ctx,
                                skip_file_matcher=False,
                                redact_secrets_and_keys=False)
        self.assertEqual([], item.getJobs())

    def test_job_source_project(self):
        base_project = model.Project('base_project', self.source)
        base_context = model.SourceContext(
            base_project.canonical_name, base_project.name,
            base_project.connection_name, 'master', 'test')
        tpc = model.TenantProjectConfig(base_project)
        tpc.trusted = True
        self.tenant.addTPC(tpc)

        base = self.pcontext.job_parser.fromYaml({
            '_source_context': base_context,
            '_start_mark': self.start_mark,
            'parent': None,
            'name': 'base',
        }, None)
        self.layout.addJob(base)

        other_project = model.Project('other_project', self.source)
        other_context = model.SourceContext(
            other_project.canonical_name, other_project.name,
            other_project.connection_name, 'master', 'test')
        tpc = model.TenantProjectConfig(other_project)
        tpc.trusted = True
        self.tenant.addTPC(tpc)
        base2 = self.pcontext.job_parser.fromYaml({
            '_source_context': other_context,
            '_start_mark': self.start_mark,
            'name': 'base',
        }, None)
        with testtools.ExpectedException(
                Exception,
                "Job base in other_project is not permitted "
                "to shadow job base in base_project"):
            self.layout.addJob(base2)

    @mock.patch("zuul.model.zkobject.ZKObject._save")
    def test_job_pipeline_allow_untrusted_secrets(self, save_mock):
        self.pipeline.post_review = False
        job = self.pcontext.job_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'job',
            'parent': None,
            'post-review': True
        }, None)

        self.layout.addJob(job)

        project_config = self.pcontext.project_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'project',
            'gate': {
                'jobs': [
                    'job'
                ]
            }
        })

        self.layout.addProjectConfig(self.project.canonical_name,
                                     project_config)

        change = model.Change(self.project)
        # Test master
        change.branch = 'master'
        change.cache_stat = Dummy(key=Dummy(reference=uuid.uuid4().hex))
        item = self.queue.enqueueChanges([change], None)
        with testtools.ExpectedException(
                model.JobConfigurationError,
                "Pre-review pipeline gate does not allow post-review job"):
            with self.zk_context as ctx:
                item.freezeJobGraph(self.layout, ctx,
                                    skip_file_matcher=False,
                                    redact_secrets_and_keys=False)

    def test_job_deduplicate_secrets(self):
        # Verify that in a job with two secrets with the same name on
        # different playbooks (achieved by inheritance) the secrets
        # are not deduplicated.  Also verify that the same secret used
        # twice is deduplicated.

        secret1_data = {'text': 'secret1 data'}
        secret2_data = {'text': 'secret2 data'}
        secret1 = self.pcontext.secret_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'secret1',
            'data': secret1_data,
        })
        self.layout.addSecret(secret1)

        secret2 = self.pcontext.secret_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'secret2',
            'data': secret2_data,
        })
        self.layout.addSecret(secret2)

        # In the first job, we test deduplication.
        base = self.pcontext.job_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'base',
            'parent': None,
            'secrets': [
                {'name': 'mysecret',
                 'secret': 'secret1'},
                {'name': 'othersecret',
                 'secret': 'secret1'},
            ],
            'pre-run': 'playbooks/pre.yaml',
        }, None)
        self.layout.addJob(base)

        # The second job should have a secret with the same name as
        # the first job, but with a different value, to make sure it
        # is not deduplicated.
        python27 = self.pcontext.job_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'python27',
            'parent': 'base',
            'secrets': [
                {'name': 'mysecret',
                 'secret': 'secret2'},
            ],
            'run': 'playbooks/python27.yaml',
        }, None)
        self.layout.addJob(python27)

        project_config = self.pcontext.project_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'project',
            'gate': {
                'jobs': ['python27'],
            }
        })
        self.layout.addProjectConfig(self.project.canonical_name,
                                     project_config)

        change = model.Change(self.project)
        change.branch = 'master'
        change.cache_stat = Dummy(key=Dummy(reference=uuid.uuid4().hex))
        item = self.queue.enqueueChanges([change], None)

        self.assertTrue(base.changeMatchesBranch(self.tenant, change))
        self.assertTrue(python27.changeMatchesBranch(self.tenant, change))

        with self.zk_context as ctx:
            item.freezeJobGraph(self.layout, ctx,
                                skip_file_matcher=False,
                                redact_secrets_and_keys=False)
        self.assertEqual(len(item.getJobs()), 1)
        job = item.getJobs()[0]
        self.assertEqual(job.name, 'python27')

        pre_idx = job.pre_run[0]['secrets']['mysecret']
        pre_secret = yaml.encrypted_load(
            job.secrets[pre_idx]['encrypted_data'])
        expected = {
            'secret_data': secret1_data,
            'secret_oidc': {},
        }
        self.assertEqual(expected, pre_secret)

        # Verify that they were deduplicated
        pre2_idx = job.pre_run[0]['secrets']['othersecret']
        self.assertEqual(pre_idx, pre2_idx)

        # Verify that the second secret is distinct
        run_idx = job.run[0]['secrets']['mysecret']
        run_secret = yaml.encrypted_load(
            job.secrets[run_idx]['encrypted_data'])
        expected = {
            'secret_data': secret2_data,
            'secret_oidc': {},
        }
        self.assertEqual(expected, run_secret)

    def _test_job_override_control(self, attr, job_attr,
                                   default, default_value,
                                   inherit, inherit_value,
                                   override, override_value,
                                   errors,
                                   ):
        # Default behavior
        data = configloader.safe_load_yaml(default, self.context)
        parent = self.pcontext.job_parser.fromYaml(data[0]['job'])
        child = self.pcontext.job_parser.fromYaml(data[1]['job'])
        job = parent.copy()
        job.setBase(self.layout, None)
        job.applyVariant(child, self.layout, None)
        self.assertEqual(default_value, getattr(job, job_attr))

        # Explicit inherit
        data = configloader.safe_load_yaml(inherit, self.context)
        parent = self.pcontext.job_parser.fromYaml(data[0]['job'])
        child = self.pcontext.job_parser.fromYaml(data[1]['job'])
        job = parent.copy()
        job.setBase(self.layout, None)
        job.applyVariant(child, self.layout, None)
        self.assertEqual(inherit_value, getattr(job, job_attr))

        # Explicit override
        data = configloader.safe_load_yaml(override, self.context)
        parent = self.pcontext.job_parser.fromYaml(data[0]['job'])
        child = self.pcontext.job_parser.fromYaml(data[1]['job'])
        job = parent.copy()
        job.setBase(self.layout, None)
        job.applyVariant(child, self.layout, None)
        self.assertEqual(override_value, getattr(job, job_attr))

        # Make sure we can't put the override in the wrong place
        for conf, exc in errors:
            with testtools.ExpectedException(exc):
                data = configloader.safe_load_yaml(conf, self.context)
                child = self.pcontext.job_parser.fromYaml(data[0]['job'])

    def _test_job_override_control_set(
            self, attr, job_attr=None,
            default_override=False,
            value_factory=lambda values: {v for v in values}):
        if job_attr is None:
            job_attr = attr
        default = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}: parent-{attr}
            - job:
                name: child
                {attr}: child-{attr}
            """)

        inherit = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}: parent-{attr}
            - job:
                name: child
                {attr}: !inherit child-{attr}
            """)
        inherit_value = value_factory([f'parent-{attr}', f'child-{attr}'])

        override = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}: parent-{attr}
            - job:
                name: child
                {attr}: !override [child-{attr}]
            """)
        override_value = value_factory([f'child-{attr}'])

        if default_override:
            default_value = override_value
        else:
            default_value = inherit_value

        errors = [(textwrap.dedent(
            f"""
            - job:
                name: child
                {attr}: [!override child-{attr}]
            """), voluptuous.error.MultipleInvalid)]
        self._test_job_override_control(attr, job_attr,
                                        default, default_value,
                                        inherit, inherit_value,
                                        override, override_value,
                                        errors)

    def test_job_override_control_tags(self):
        self._test_job_override_control_set('tags')

    def test_job_override_control_requires(self):
        self._test_job_override_control_set('requires')

    def test_job_override_control_provides(self):
        self._test_job_override_control_set('provides')

    def test_job_override_control_include_vars(self):
        self._test_job_override_control_set(
            'include-vars',
            job_attr='include_vars',
            value_factory=lambda values:
                [model.JobIncludeVars(v, 'git.example.com/project', True, True)
                 for v in values]
        )

    def test_job_override_control_dependencies(self):
        self._test_job_override_control_set(
            'dependencies',
            default_override=True,
            value_factory=lambda values:
            {model.JobDependency(v) for v in values})

    def test_job_override_control_failure_output(self):
        self._test_job_override_control_set(
            'failure-output',
            job_attr='failure_output',
            value_factory=lambda values: tuple(v for v in sorted(values)))

    def test_job_override_control_files(self):
        self._test_job_override_control_set(
            'files',
            job_attr='file_matcher',
            default_override=True,
            value_factory=lambda values: change_matcher.MatchAnyFiles(
                [change_matcher.FileMatcher(ZuulRegex(v))
                 for v in sorted(values)]))

    def test_job_override_control_irrelevant_files(self):
        self._test_job_override_control_set(
            'irrelevant-files',
            job_attr='irrelevant_file_matcher',
            default_override=True,
            value_factory=lambda values: change_matcher.MatchAllFiles(
                [change_matcher.FileMatcher(ZuulRegex(v))
                 for v in sorted(values)]))

    def _test_job_override_control_dict(
            self, attr, job_attr=None,
            default_override=False):
        if job_attr is None:
            job_attr = attr
        default = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}:
                  parent: 1
            - job:
                name: child
                {attr}:
                  child: 2
            """)

        inherit = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}:
                   parent: 1
            - job:
                name: child
                {attr}: !inherit
                   child: 2
            """)
        inherit_value = {'parent': 1, 'child': 2}

        override = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}:
                   parent: 1
            - job:
                name: child
                {attr}: !override
                   child: 2
            """)
        override_value = {'child': 2}

        if default_override:
            default_value = override_value
        else:
            default_value = inherit_value

        errors = [
            (textwrap.dedent(
                f"""
                - job:
                    name: child
                    {attr}:
                      !override child: 2
                """), voluptuous.error.MultipleInvalid),
            (textwrap.dedent(
                f"""
                - job:
                    name: child
                    {attr}:
                      child: !override 2
                """), Exception),
        ]
        self._test_job_override_control(attr, job_attr,
                                        default, default_value,
                                        inherit, inherit_value,
                                        override, override_value,
                                        errors)

    def test_job_override_control_vars(self):
        self._test_job_override_control_dict(
            'vars', job_attr='variables')

    def test_job_override_control_extra_vars(self):
        self._test_job_override_control_dict(
            'extra-vars', job_attr='extra_variables')

    def _test_job_override_control_host_dict(
            self, attr, job_attr=None,
            default_override=False):
        if job_attr is None:
            job_attr = attr
        default = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}:
                  host:
                    parent: 1
            - job:
                name: child
                {attr}:
                  host:
                    child: 2
            """)

        inherit = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}:
                   host:
                     parent: 1
            - job:
                name: child
                {attr}: !inherit
                   host:
                     child: 2
            """)
        inherit_value = {'host': {'parent': 1, 'child': 2}}

        override = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}:
                   host:
                     parent: 1
            - job:
                name: child
                {attr}: !override
                   host:
                     child: 2
            """)
        override_value = {'host': {'child': 2}}

        if default_override:
            default_value = override_value
        else:
            default_value = inherit_value

        errors = [
            (textwrap.dedent(
                f"""
                - job:
                    name: child
                    {attr}:
                      !override host:
                        child: 2
                """), voluptuous.error.MultipleInvalid),
            (textwrap.dedent(
                f"""
                - job:
                    name: child
                    {attr}:
                      host: !override
                        child: 2
                """), voluptuous.error.MultipleInvalid),
            (textwrap.dedent(
                f"""
                - job:
                    name: child
                    {attr}:
                      host:
                        child: !override 2
            """), Exception),
        ]
        self._test_job_override_control(attr, job_attr,
                                        default, default_value,
                                        inherit, inherit_value,
                                        override, override_value,
                                        errors)

    def test_job_override_control_host_vars(self):
        self._test_job_override_control_host_dict(
            'host-vars', job_attr='host_variables')

    def test_job_override_control_group_vars(self):
        self._test_job_override_control_host_dict(
            'group-vars', job_attr='group_variables')

    def test_job_override_control_required_projects(self):
        parent = model.Project('parent-project', self.source)
        child = model.Project('child-project', self.source)
        parent_tpc = model.TenantProjectConfig(parent)
        child_tpc = model.TenantProjectConfig(child)
        self.tenant.addTPC(parent_tpc)
        self.tenant.addTPC(child_tpc)

        default = textwrap.dedent(
            """
            - job:
                name: parent
                required-projects: parent-project
            - job:
                name: child
                required-projects: child-project
            """)

        inherit = textwrap.dedent(
            """
            - job:
                name: parent
                required-projects: parent-project
            - job:
                name: child
                required-projects: !inherit child-project
            """)
        inherit_value = {
            'git.example.com/parent-project': model.JobProject(
                'git.example.com/parent-project'),
            'git.example.com/child-project': model.JobProject(
                'git.example.com/child-project'),
        }

        override = textwrap.dedent(
            """
            - job:
                name: parent
                required-projects: parent-project
            - job:
                name: child
                required-projects: !override [child-project]
            """)
        override_value = {
            'git.example.com/child-project': model.JobProject(
                'git.example.com/child-project'),
        }

        default_value = inherit_value

        errors = [(textwrap.dedent(
            """
            - job:
                name: child
                required-projects: [!override child-project]
            """), voluptuous.error.MultipleInvalid)]
        self._test_job_override_control('required-projects',
                                        'required_projects',
                                        default, default_value,
                                        inherit, inherit_value,
                                        override, override_value,
                                        errors)

    def _test_job_final_control(self, attr, job_attr,
                                default, default_value,
                                final):
        # Default behavior
        data = configloader.safe_load_yaml(default, self.context)
        parent = self.pcontext.job_parser.fromYaml(data[0]['job'])
        child = self.pcontext.job_parser.fromYaml(data[1]['job'])
        job = parent.copy()
        job.setBase(self.layout, None)
        job.applyVariant(child, self.layout, None)
        self.assertEqual(default_value, getattr(job, job_attr))

        # Verify final attr exception
        data = configloader.safe_load_yaml(final, self.context)
        parent = self.pcontext.job_parser.fromYaml(data[0]['job'])
        child = self.pcontext.job_parser.fromYaml(data[1]['job'])
        job = parent.copy()
        job.setBase(self.layout, None)
        with testtools.ExpectedException(model.JobConfigurationError,
                                         ".* final attribute"):
            job.applyVariant(child, self.layout, None)

    def _test_job_final_control_set(
            self, attr, job_attr=None,
            default_override=False,
            value_factory=lambda values: {v for v in values}):
        if job_attr is None:
            job_attr = attr
        default = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}: parent-{attr}
            - job:
                name: child
                {attr}: child-{attr}
            """)
        inherit_value = value_factory([f'parent-{attr}', f'child-{attr}'])
        override_value = value_factory([f'child-{attr}'])
        if default_override:
            default_value = override_value
        else:
            default_value = inherit_value

        final = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}: parent-{attr}
                attribute-control:
                  {attr}:
                    final: true
            - job:
                name: child
                {attr}: child-{attr}
            """)

        self._test_job_final_control(attr, job_attr,
                                     default, default_value,
                                     final)

    def test_job_final_control_tags(self):
        self._test_job_final_control_set('tags')

    def test_job_final_control_requires(self):
        self._test_job_final_control_set('requires')

    def test_job_final_control_provides(self):
        self._test_job_final_control_set('provides')

    def test_job_final_control_dependencies(self):
        self._test_job_final_control_set(
            'dependencies',
            default_override=True,
            value_factory=lambda values:
            {model.JobDependency(v) for v in values})

    def test_job_final_control_failure_output(self):
        self._test_job_final_control_set(
            'failure-output',
            job_attr='failure_output',
            value_factory=lambda values: tuple(v for v in sorted(values)))

    def test_job_final_control_files(self):
        self._test_job_final_control_set(
            'files',
            job_attr='file_matcher',
            default_override=True,
            value_factory=lambda values: change_matcher.MatchAnyFiles(
                [change_matcher.FileMatcher(ZuulRegex(v))
                 for v in sorted(values)]))

    def test_job_final_control_irrelevant_files(self):
        self._test_job_final_control_set(
            'irrelevant-files',
            job_attr='irrelevant_file_matcher',
            default_override=True,
            value_factory=lambda values: change_matcher.MatchAllFiles(
                [change_matcher.FileMatcher(ZuulRegex(v))
                 for v in sorted(values)]))

    def _test_job_final_control_dict(
            self, attr, job_attr=None,
            default_override=False):
        if job_attr is None:
            job_attr = attr
        default = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}:
                  parent: 1
            - job:
                name: child
                {attr}:
                  child: 2
            """)

        final = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}:
                   parent: 1
                attribute-control:
                  {attr}:
                    final: true
            - job:
                name: child
                {attr}:
                   child: 2
            """)
        inherit_value = {'parent': 1, 'child': 2}
        default_value = {'child': 2}

        if default_override:
            default_value = default_value
        else:
            default_value = inherit_value

        self._test_job_final_control(attr, job_attr,
                                     default, default_value,
                                     final)

    def test_job_final_control_vars(self):
        self._test_job_final_control_dict(
            'vars', job_attr='variables')

    def test_job_final_control_extra_vars(self):
        self._test_job_final_control_dict(
            'extra-vars', job_attr='extra_variables')

    def _test_job_final_control_host_dict(
            self, attr, job_attr=None,
            default_override=False):
        if job_attr is None:
            job_attr = attr
        default = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}:
                  host:
                    parent: 1
            - job:
                name: child
                {attr}:
                  host:
                    child: 2
            """)

        final = textwrap.dedent(
            f"""
            - job:
                name: parent
                {attr}:
                   host:
                     parent: 1
                attribute-control:
                  {attr}:
                    final: true
            - job:
                name: child
                {attr}:
                   host:
                     child: 2
            """)
        inherit_value = {'host': {'parent': 1, 'child': 2}}
        default_value = {'host': {'child': 2}}

        if default_override:
            default_value = default_value
        else:
            default_value = inherit_value

        self._test_job_final_control(attr, job_attr,
                                     default, default_value,
                                     final)

    def test_job_final_control_host_vars(self):
        self._test_job_final_control_host_dict(
            'host-vars', job_attr='host_variables')

    def test_job_final_control_group_vars(self):
        self._test_job_final_control_host_dict(
            'group-vars', job_attr='group_variables')

    def test_job_final_control_required_projects(self):
        parent = model.Project('parent-project', self.source)
        child = model.Project('child-project', self.source)
        parent_tpc = model.TenantProjectConfig(parent)
        child_tpc = model.TenantProjectConfig(child)
        self.tenant.addTPC(parent_tpc)
        self.tenant.addTPC(child_tpc)

        default = textwrap.dedent(
            """
            - job:
                name: parent
                required-projects: parent-project
            - job:
                name: child
                required-projects: child-project
            """)

        final = textwrap.dedent(
            """
            - job:
                name: parent
                required-projects: parent-project
                attribute-control:
                  required-projects:
                    final: true
            - job:
                name: child
                required-projects: child-project
            """)

        default_value = {
            'git.example.com/parent-project': model.JobProject(
                'git.example.com/parent-project'),
            'git.example.com/child-project': model.JobProject(
                'git.example.com/child-project'),
        }

        self._test_job_final_control('required-projects',
                                     'required_projects',
                                     default, default_value,
                                     final)

    def test_job_final_control_type(self):
        default = textwrap.dedent(
            """
            - job:
                name: parent
                type: initializer
            - job:
                name: child
                type: regular
            """)

        final = textwrap.dedent(
            """
            - job:
                name: parent
                type: initializer
                attribute-control:
                  type:
                    final: true
            - job:
                name: child
                type: regular
            """)

        default_value = 'regular'

        self._test_job_final_control('type',
                                     'type',
                                     default, default_value,
                                     final)

    @mock.patch("zuul.model.zkobject.ZKObject._save")
    def test_image_permissions(self, save_mock):
        self.pipeline.post_review = False
        image = self.pcontext.image_parser.fromYaml({
            '_source_context': self.context,
            '_start_mark': self.start_mark,
            'name': 'image',
            'type': 'zuul',
        })
        self.layout.addImage(image)

        job_project = model.Project('other-project', self.source)
        job_context = model.SourceContext(
            job_project.canonical_name, job_project.name,
            job_project.connection_name, 'master', 'test')
        tpc = model.TenantProjectConfig(job_project)
        tpc.trusted = True
        self.tenant.addTPC(tpc)

        job = self.pcontext.job_parser.fromYaml({
            '_source_context': job_context,
            '_start_mark': self.start_mark,
            'name': 'job',
            'parent': None,
            'post-review': True,
            'image-build-name': 'image',
        }, None)

        self.layout.addJob(job)

        project_config = self.pcontext.project_parser.fromYaml({
            '_source_context': job_context,
            '_start_mark': self.start_mark,
            'name': 'other-project',
            'gate': {
                'jobs': [
                    'job'
                ]
            }
        })

        self.layout.addProjectConfig(job_project.canonical_name,
                                     project_config)

        change = model.Change(job_project)
        # Test master
        change.branch = 'master'
        change.cache_stat = Dummy(key=Dummy(reference=uuid.uuid4().hex))
        item = self.queue.enqueueChanges([change], None)
        with testtools.ExpectedException(
                model.JobConfigurationError,
                "The image build job"):
            with self.zk_context as ctx:
                item.freezeJobGraph(self.layout, ctx,
                                    skip_file_matcher=False,
                                    redact_secrets_and_keys=False)


class FakeFrozenJob(model.Job):

    def __init__(self, name):
        super().__init__(name)
        self.uuid = uuid.uuid4().hex
        self.ref = 'fake reference'
        self.all_refs = [self.ref]
        self.matches_change = True

    def _set(self, **kw):
        self.__dict__.update(kw)


class TestGraph(BaseTestCase):
    def setUp(self):
        COMPONENT_REGISTRY.registry = Dummy()
        COMPONENT_REGISTRY.registry.model_api = MODEL_API
        super().setUp()

    def test_job_graph_disallows_circular_dependencies(self):
        jobs = [FakeFrozenJob('job%d' % i) for i in range(0, 10)]

        def setup_graph():
            graph = model.JobGraph({})
            prevjob = None
            for j in jobs[:3]:
                if prevjob:
                    j.dependencies = frozenset([
                        model.JobDependency(prevjob.name)])
                graph.addJob(j)
                prevjob = j
            # 0 triggers 1 triggers 2 triggers 3...
            return graph

        # Cannot depend on itself
        graph = setup_graph()
        j = FakeFrozenJob('jobX')
        j.dependencies = frozenset([model.JobDependency(j.name)])
        graph.addJob(j)
        with testtools.ExpectedException(
                Exception,
                "Dependency cycle detected in job jobX"):
            graph.freezeDependencies(self.log)

        # Disallow circular dependencies
        graph = setup_graph()
        jobs[3].dependencies = frozenset([model.JobDependency(jobs[4].name)])
        graph.addJob(jobs[3])
        jobs[4].dependencies = frozenset([model.JobDependency(jobs[3].name)])
        graph.addJob(jobs[4])
        with testtools.ExpectedException(
                Exception,
                "Dependency cycle detected in job job3"):
            graph.freezeDependencies(self.log)

        graph = setup_graph()
        jobs[3].dependencies = frozenset([model.JobDependency(jobs[5].name)])
        graph.addJob(jobs[3])
        jobs[4].dependencies = frozenset([model.JobDependency(jobs[3].name)])
        graph.addJob(jobs[4])
        jobs[5].dependencies = frozenset([model.JobDependency(jobs[4].name)])
        graph.addJob(jobs[5])

        with testtools.ExpectedException(
                Exception,
                "Dependency cycle detected in job job3"):
            graph.freezeDependencies(self.log)

        graph = setup_graph()
        jobs[3].dependencies = frozenset([model.JobDependency(jobs[2].name)])
        graph.addJob(jobs[3])
        jobs[4].dependencies = frozenset([model.JobDependency(jobs[3].name)])
        graph.addJob(jobs[4])
        jobs[5].dependencies = frozenset([model.JobDependency(jobs[4].name)])
        graph.addJob(jobs[5])
        jobs[6].dependencies = frozenset([model.JobDependency(jobs[2].name)])
        graph.addJob(jobs[6])
        graph.freezeDependencies(self.log)

    def test_job_graph_allows_soft_dependencies(self):
        parent = FakeFrozenJob('parent')
        child = FakeFrozenJob('child')
        child.dependencies = frozenset([
            model.JobDependency(parent.name, True)])

        # With the parent
        graph = model.JobGraph({})
        graph.addJob(parent)
        graph.addJob(child)
        graph.freezeDependencies(self.log)
        self.assertEqual(graph.getParentJobsRecursively(child),
                         [parent])

        # Skip the parent
        graph = model.JobGraph({})
        graph.addJob(child)
        graph.freezeDependencies(self.log)
        self.assertEqual(graph.getParentJobsRecursively(child), [])

    def test_job_graph_allows_soft_dependencies4(self):
        # A more complex scenario with multiple parents at each level
        parents = [FakeFrozenJob('parent%i' % i) for i in range(6)]
        child = FakeFrozenJob('child')
        child.dependencies = frozenset([
            model.JobDependency(parents[0].name, True),
            model.JobDependency(parents[1].name)])
        parents[0].dependencies = frozenset([
            model.JobDependency(parents[2].name),
            model.JobDependency(parents[3].name, True)])
        parents[1].dependencies = frozenset([
            model.JobDependency(parents[4].name),
            model.JobDependency(parents[5].name)])
        # Run them all
        graph = model.JobGraph({})
        for j in parents:
            graph.addJob(j)
        graph.addJob(child)
        graph.freezeDependencies(self.log)
        self.assertEqual(set(graph.getParentJobsRecursively(child)),
                         set(parents))

        # Skip first parent, therefore its recursive dependencies don't appear
        graph = model.JobGraph({})
        for j in parents:
            if j is not parents[0]:
                graph.addJob(j)
        graph.addJob(child)
        graph.freezeDependencies(self.log)
        self.assertEqual(set(graph.getParentJobsRecursively(child)),
                         set(parents) -
                         set([parents[0], parents[2], parents[3]]))

        # Skip a leaf node
        graph = model.JobGraph({})
        for j in parents:
            if j is not parents[3]:
                graph.addJob(j)
        graph.addJob(child)
        graph.freezeDependencies(self.log)
        self.assertEqual(set(graph.getParentJobsRecursively(child)),
                         set(parents) - set([parents[3]]))

    def test_soft_dependency_mixed_cycle(self):
        # This is a regression test to make sure we are not processing
        # jobs twice, when mixing soft/hard dependencies to a job.
        graph = model.JobGraph({})
        parent = FakeFrozenJob("parent")
        graph.addJob(parent)

        intermediate = FakeFrozenJob("intermediate")
        intermediate.dependencies = frozenset([
            # Hard dependency to parent
            model.JobDependency(parent.name)
        ])
        graph.addJob(intermediate)

        child = FakeFrozenJob("child")
        child.dependencies = frozenset([
            # Soft dependency to parent
            model.JobDependency(parent.name, soft=True),
            # Dependency to intermediate, which has a hard dependency
            # to parent.
            model.JobDependency(intermediate.name),
        ])
        graph.addJob(child)

        # We don't expect this to raise an exception
        graph.freezeDependencies(self.log)


class TestTenant(BaseTestCase):
    def test_add_project(self):
        tenant = model.Tenant('tenant')
        connection1 = Dummy(connection_name='dummy_connection1')
        source1 = Dummy(canonical_hostname='git1.example.com',
                        name='dummy',  # TODOv3(jeblair): remove
                        connection=connection1)

        source1_project1 = model.Project('project1', source1)
        source1_project1_tpc = model.TenantProjectConfig(source1_project1)
        source1_project1_tpc.trusted = True
        tenant.addTPC(source1_project1_tpc)
        d = {'project1':
             {'git1.example.com': source1_project1}}
        self.assertEqual(d, tenant.projects)
        self.assertEqual((True, source1_project1),
                         tenant.getProject('project1'))
        self.assertEqual((True, source1_project1),
                         tenant.getProject('git1.example.com/project1'))

        source1_project2 = model.Project('project2', source1)
        tpc = model.TenantProjectConfig(source1_project2)
        tenant.addTPC(tpc)
        d = {'project1':
             {'git1.example.com': source1_project1},
             'project2':
             {'git1.example.com': source1_project2}}
        self.assertEqual(d, tenant.projects)
        self.assertEqual((False, source1_project2),
                         tenant.getProject('project2'))
        self.assertEqual((False, source1_project2),
                         tenant.getProject('git1.example.com/project2'))

        connection2 = Dummy(connection_name='dummy_connection2')
        source2 = Dummy(canonical_hostname='git2.example.com',
                        name='dummy',  # TODOv3(jeblair): remove
                        connection=connection2)

        source2_project1 = model.Project('project1', source2)
        tpc = model.TenantProjectConfig(source2_project1)
        tenant.addTPC(tpc)
        d = {'project1':
             {'git1.example.com': source1_project1,
              'git2.example.com': source2_project1},
             'project2':
             {'git1.example.com': source1_project2}}
        self.assertEqual(d, tenant.projects)
        with testtools.ExpectedException(
                Exception,
                "Project name 'project1' is ambiguous"):
            tenant.getProject('project1')
        self.assertEqual((False, source1_project2),
                         tenant.getProject('project2'))
        self.assertEqual((True, source1_project1),
                         tenant.getProject('git1.example.com/project1'))
        self.assertEqual((False, source2_project1),
                         tenant.getProject('git2.example.com/project1'))

        source2_project2 = model.Project('project2', source2)
        tpc = model.TenantProjectConfig(source2_project2)
        tpc.trusted = True
        tenant.addTPC(tpc)
        d = {'project1':
             {'git1.example.com': source1_project1,
              'git2.example.com': source2_project1},
             'project2':
             {'git1.example.com': source1_project2,
              'git2.example.com': source2_project2}}
        self.assertEqual(d, tenant.projects)
        with testtools.ExpectedException(
                Exception,
                "Project name 'project1' is ambiguous"):
            tenant.getProject('project1')
        with testtools.ExpectedException(
                Exception,
                "Project name 'project2' is ambiguous"):
            tenant.getProject('project2')
        self.assertEqual((True, source1_project1),
                         tenant.getProject('git1.example.com/project1'))
        self.assertEqual((False, source2_project1),
                         tenant.getProject('git2.example.com/project1'))
        self.assertEqual((False, source1_project2),
                         tenant.getProject('git1.example.com/project2'))
        self.assertEqual((True, source2_project2),
                         tenant.getProject('git2.example.com/project2'))

        source1_project2b = model.Project('subpath/project2', source1)
        tpc = model.TenantProjectConfig(source1_project2b)
        tpc.trusted = True
        tenant.addTPC(tpc)
        d = {'project1':
             {'git1.example.com': source1_project1,
              'git2.example.com': source2_project1},
             'project2':
             {'git1.example.com': source1_project2,
              'git2.example.com': source2_project2},
             'subpath/project2':
             {'git1.example.com': source1_project2b}}
        self.assertEqual(d, tenant.projects)
        self.assertEqual((False, source1_project2),
                         tenant.getProject('git1.example.com/project2'))
        self.assertEqual((True, source2_project2),
                         tenant.getProject('git2.example.com/project2'))
        self.assertEqual((True, source1_project2b),
                         tenant.getProject('subpath/project2'))
        self.assertEqual(
            (True, source1_project2b),
            tenant.getProject('git1.example.com/subpath/project2'))

        source2_project2b = model.Project('subpath/project2', source2)
        tpc = model.TenantProjectConfig(source2_project2b)
        tpc.trusted = True
        tenant.addTPC(tpc)
        d = {'project1':
             {'git1.example.com': source1_project1,
              'git2.example.com': source2_project1},
             'project2':
             {'git1.example.com': source1_project2,
              'git2.example.com': source2_project2},
             'subpath/project2':
             {'git1.example.com': source1_project2b,
              'git2.example.com': source2_project2b}}
        self.assertEqual(d, tenant.projects)
        self.assertEqual((False, source1_project2),
                         tenant.getProject('git1.example.com/project2'))
        self.assertEqual((True, source2_project2),
                         tenant.getProject('git2.example.com/project2'))
        with testtools.ExpectedException(
                Exception,
                "Project name 'subpath/project2' is ambiguous"):
            tenant.getProject('subpath/project2')
        self.assertEqual(
            (True, source1_project2b),
            tenant.getProject('git1.example.com/subpath/project2'))
        self.assertEqual(
            (True, source2_project2b),
            tenant.getProject('git2.example.com/subpath/project2'))

        with testtools.ExpectedException(
                Exception,
                "Project project1 is already in project index"):
            tenant._addProject(source1_project1_tpc)


class TestRef(BaseTestCase):
    def test_ref_equality(self):
        change1 = model.Change('project1')
        change1.ref = '/change1'
        change1b = model.Change('project1')
        change1b.ref = '/change1'
        change2 = model.Change('project2')
        change2.ref = '/change2'
        self.assertFalse(change1.equals(change2))
        self.assertTrue(change1.equals(change1b))

        tag1 = model.Tag('project1')
        tag1.ref = '/tag1'
        tag1b = model.Tag('project1')
        tag1b.ref = '/tag1'
        tag2 = model.Tag('project2')
        tag2.ref = '/tag2'
        self.assertFalse(tag1.equals(tag2))
        self.assertTrue(tag1.equals(tag1b))

        self.assertFalse(tag1.equals(change1))

        branch1 = model.Branch('project1')
        branch1.ref = '/branch1'
        branch1b = model.Branch('project1')
        branch1b.ref = '/branch1'
        branch2 = model.Branch('project2')
        branch2.ref = '/branch2'
        self.assertFalse(branch1.equals(branch2))
        self.assertTrue(branch1.equals(branch1b))

        self.assertFalse(branch1.equals(change1))
        self.assertFalse(branch1.equals(tag1))


class TestSourceContext(BaseTestCase):
    def setUp(self):
        super().setUp()
        COMPONENT_REGISTRY.registry = Dummy()
        COMPONENT_REGISTRY.registry.model_api = MODEL_API
        self.connection = Dummy(connection_name='dummy_connection')
        self.source = Dummy(canonical_hostname='git.example.com',
                            connection=self.connection)
        self.project = model.Project('project', self.source)
        self.context = model.SourceContext(
            self.project.canonical_name, self.project.name,
            self.project.connection_name, 'master', 'test')
        self.context.implied_branches = [
            change_matcher.BranchMatcher(ZuulRegex('foo')),
            change_matcher.ImpliedBranchMatcher(ZuulRegex('foo')),
        ]

    def test_serialize(self):
        self.context.deserialize(self.context.serialize())
